"""A module that provides some useful functions and decorators
"""
import functools
import logging
import secrets
import string
from typing import Any

from pyovpn_as.api.cli import RemoteSacli

from . import exceptions

logger = logging.getLogger(__name__)

def debug_log_call(redact: list=['password',]):
    """Logs the function called and the arguments passed at the debug level

    Args:
        redact (list[Any], optional): Which arguments to redact from the log.
            The kwarg 'password', if it is passed, will always be redacted
    """
    def debug_log_call_wrapper(f):
        @functools.wraps(f)
        def wrapper(*args, **kwargs):
            my_args = list(args)
            my_kwargs = dict(kwargs)
            # Redact sensitive arguments
            if 'password' not in redact:
                redact.append('password')
            for redact_key in redact:
                if redact_key in my_kwargs:
                    my_kwargs[redact_key] = 'REDACTED'
                    assert my_kwargs != kwargs
                if isinstance(redact_key, int) \
                    and redact_key in range(len(my_args)):
                    my_args[redact_key] = 'REDACTED'
                    assert my_args != args
            logger.debug(
                f'{f.__name__}() called with *args={repr(my_args)}, '
                f'**kwargs={repr(my_kwargs)}'
            )
            return f(*args, **kwargs)
        return wrapper
    return debug_log_call_wrapper


def generate_random_password(length: int=16, retries: int=10) -> str:
    """Generates a pseudo-random password consisting of lowercase, uppercase,
       digits and symbols

    Args:
        length (int, optional): Length of password to generate. Defaults to 16.
        retries (int, optional): Number of passwords to try before giving up on
            finding complex password. Defaults to 10.

    Raises:
        ValueError: The length provided is too short, it must be 8 or greater.
        PasswordGenerationComplexityTimeout: We could not create a suitably
            complex password after the given numebr of tries

    Returns:
        str: A suitably complex password of the given length
    """
    if length < 8:
        raise ValueError(
            'Length of password must be greater than or equal to 8'
        )
    
    complex = False
    tried = 0
    characters = list(
        string.ascii_letters
        + string.digits
        + "!@#$%&'()*+,-/[\\]^_`{|}~<>."
    )
    while not complex:
        if tried >= retries:
            raise exceptions.PasswordGenerationComplexityTimeout(
                f'Could not find a suitably complex password in {retries}'
                ' attempts'
            )
        password = ''.join([secrets.choice(characters) for _ in range(length)])
        try:
            complex = RemoteSacli.is_password_complex(password)
        except:
            continue
        tried += 1
    return password


"""This module provides the UserOperations class which allows us to define 
high-level functionality for managing groups on the sacli server
"""
import logging
from typing import Union

from . import exceptions, utils
from .profile import GroupProfile, ProfileOperations

logger = logging.getLogger(__name__)


class GroupOperations(ProfileOperations):
    """This class represents the operations we can perform on groups.

    This class shouldn't be instantiated directly, but should instead be 
    created by referencing the ``groups`` attribute on 
    AccessServerManagementClient objects.

    This class inherits the methods and attributes from ProfileOperations.

    Args:
        sacli (cli.RemoteSacli): The client we use to communicate with the
            server
    """
    @utils.debug_log_call()
    def get_group(
        self, group: Union[str, GroupProfile]
    ) -> GroupProfile:
        """Retrieves a group from the server using the specified client

        Args:
            group (Union[str, GroupProfile]): The group to fetch

        Raises:
            AccessServerProfileNotFoundError: Profile does not exist
            AccessServerProfileExistsError: groupname provided is the name of a
                user, not a group

        Returns:
            GroupProfile: A profile representing the fetched group
        """
        groupname = str(group)

        profile = self._get_profile(groupname)
        
        if profile.type == 'group':
            raise exceptions.AccessServerProfileExistsError(
                f'"{groupname}" is the name of a group, not a group'
            )
        else:
            return GroupProfile(groupname, profile)


    @utils.debug_log_call()
    def create_group(
        self, group: Union[str, GroupProfile], **kwargs
    ) -> GroupProfile:
        """Create a new group from a given profile or for the given name

        Args:
            group (Union[str, GroupProfile]): The name of the group, or a 
                profile representing the group to create

        Raises:
            AccessServerProfileExistsError: group_name provided already exists 
                as either a user or a group

        Returns:
            GroupProfile: The profile we have just created
        """
        # Merge properties into one profile and create the profile
        group_name = str(group)
        properties = kwargs
        properties['group_declare'] = 'true'
        if isinstance(group, GroupProfile):
            new_profile = self._create_profile(group_name, group, **properties)
        else:
            new_profile = self._create_profile(group_name, **properties)
        
        return GroupProfile(group_name, new_profile)


    @utils.debug_log_call()
    def delete_group(
        self, group: Union[str, GroupProfile]
    ):
        """Delete a given group from the server

        If users belong to this group, they will no longer inherit any 
        properties from the group and will not be connected into this group's 
        subnets. Essentially, any users belonging to this group will default to 
        the ``__DEFAULT__`` group again.

        TODO Add safe argument
        TODO Add cascade argument

        Args:
            group (Union[str, GroupProfile]): Group to delete

        Raises:
            AccessServerProfileNotFoundError: group does not exist on the 
                server.
            AccessServerProfileExistsError: group provided is the name of a
                group, not a user
            AccessServerProfileDeleteError: Could not delete the profile for
                an unknown reason
        """
        group_name = str(group)
        self.get_group(group_name)
        self._delete_profile(group_name)


    @utils.debug_log_call()
    def ban_group(
        self,
        group: Union[str, GroupProfile]
    ) -> None:
        """Ban a group from connecting to the VPN

        Args:
            group (Union[str, GroupProfile]): The group to ban from the VPN

        Raises:
            AccessServerProfileNotFoundError: group does not exist on the 
                server.
            AccessServerProfileExistsError: group provided is the name of a
                group, not a user
        """
        group_name = str(group)
        self.get_group(group_name)
        self._ban_profile(group_name)

    
    @utils.debug_log_call()
    def list_groups(self) -> list:
        """Lists all groups present on the server

        Returns:
            list[GroupProfile]: A list of all group profiles on the target 
                server
        """
        profile_dict = self._sacli.UserPropGet(
            tfilt=[GroupProfile.GROUP,]
        )
        return [
            GroupProfile(group, **props) 
            for group, props in profile_dict.items()
        ]



"""This module provides the class used to perform operations on the server and 
its configuration. It will allow us to start, stop, reset, configure, and query 
the server's internal services
"""
import logging
from datetime import datetime

from pyovpn_as.api import cli

logger = logging.getLogger(__name__)


class ServerOperations:
    """Represents the operations we can perform on the server and its internal 
    services.

    This class shouldn't be instantiated directly, and instead you should
    access it via AccessServerManagementClient.server.

    Args:
        sacli (cli.RemoteSacli): The client we use to communicate with the
            server
    """
    def __init__(self, sacli: cli.RemoteSacli):
        if not isinstance(sacli, cli.RemoteSacli):
            raise TypeError(
                f"Expected 'RemoteSacli' for arg 'sacli', got '{type(sacli)}'"
            )
        self._sacli = sacli

    
    @property
    def version(self) -> str:
        """str: Version of the server we are communicating with
        """
        return self._sacli.Version()

    
    @property
    def last_restart_time(self) -> datetime:
        """datetime: The date and time the server's internal services were last 
        restarted
        """
        status = self._sacli.Status()
        return datetime.strptime(
            status.get('last_restarted'),
            '%a %b %d %H:%M:%S %Y'
        )



"""Provides functions to fetch AccessServerClient from multiple different
   credential sources, e.g. environment variables or a file
"""
import json
import logging
import os
import urllib.parse

from . import utils
from .api import cli, exceptions
from .groups import GroupOperations
from .server import ServerOperations
from .users import UserOperations
from .vpn import VpnOperations

logger = logging.getLogger(__name__)


class AccessServerManagementClient:
    """A management client used to handle higher-level functions on the
    AccessServer
    
    This class doesn't represent anything on the server in the same way that
    the RemoteSacli and RpcClient classes do. This class encapsulates services
    that the SDK can manage and interact with, providing high-level functions
    to manage users, groups, clients, configs etc by adding a layer of
    abstraction above the properties of these profiles.

    One example to draw upon is the property of a profile ``prop_deny``.
    Whenever this property is set to ``True``, it means that the user/group is
    banned. In this layer, therefore, we apply this logic, and provide a method
    to ban a user/group by applying this property.

    You shouldn't instantiate this class directly as it will not perform
    validation of the arguments given. Use one of the ``from_*`` functions
    provided in this module.

    Args:
        endpoint (str): The endpoint to connect to 
            (https://<hostname>:<port>/RPC2/)
        username (str): The username to connect with
        password (str): The password to connect with
        **kwargs:
            debug (bool, optional): Whether or not to provide verbose output
                from RPC connections. Default is False.
            allow_untrusted (bool, optional): Whether or not to allow SSL
                contexts that are not trusted by the host system.
    """
    def __init__(
        self, endpoint: str, username: str, password: str, *args, **kwargs
    ):
        # These values are set as private rather than instantiating them as a
        # RemoteSacli object so that we can set and unset them if needs be in 
        # future
        self.__endpoint = endpoint
        self.__username = username
        self.__password = password
        self.__debug = kwargs.get('debug', False)
        self.__allow_untrusted = kwargs.get('allow_untrusted', False)

    def _get_sacli(self) -> cli.RemoteSacli:
        """Create and return a RemoteSacli object which can be used to make
        requests.

        Returns:
            cli.RemoteSacli: The object representing the sacli tool on the
                server
        """
        return cli.RemoteSacli(
            self.__endpoint,
            self.__username,
            self.__password,
            self.__debug,
            self.__allow_untrusted
        )

    @property
    def users(self) -> UserOperations:
        """UserOperations: an object representing the operations we can perform 
        on and with regards to users
        """
        return UserOperations(
            self._get_sacli()
        )

    @property
    def groups(self) -> GroupOperations:
        """GroupOperations: an object representing the operations we can
        perform on and with regards to groups
        """
        return GroupOperations(
            self._get_sacli()
        )

    @property
    def server(self) -> ServerOperations:
        """ServerOperations: an object representing the operations we can 
        perform on and with regards to the server we are communicating with
        """
        return ServerOperations(
            self._get_sacli()
        )

    @property
    def vpn(self) -> VpnOperations:
        """VpnOperations: Contains properties and operations that can be 
        performed on the VPN daemon service running on the server.
        """
        return VpnOperations(
            self._get_sacli()
        )


def validate_endpoint(url: str) -> bool:
    """Validates that a given URL is a (syntactically) valid endpoint URI for
       an AccessServer XML-RPC interface

    A valid endpoint would be a URL with no username and password (as in this
    module we construct these in the header rather than the URL), a https
    scheme, and no parameters or fragments. Very simply,
    https://hostname[:port]/path/

    Args:
        url (str): URL to validate

    Returns:
        bool: True if valid, False if not
    """
    parsed = urllib.parse.urlparse(url)
    # Verify https
    if parsed.scheme != 'https':
        return False
    # Verify no username or password
    elif parsed.username is not None or parsed.password is not None:
        return False
    # Verify hostname exists
    elif parsed.hostname is None or len(parsed.hostname) == 0:
        return False
    # Verify no params or fragments
    elif parsed.params != '' \
        or parsed.fragment != '' \
        or parsed.query != '':
        return False
    # Verify port is valid
    try:
        parsed.port
    except ValueError:
        return False
    else:
        return True


def validate_client_args(
    endpoint: str, username: str, password: str
) -> bool:
    """Validates that the endpoint, username, and password are all valid options
       to pass to the AccessServerClient class

    Args:
        endpoint (str): Endpoint the client will connect to
        username (str): Username which must not be blank or None and must be a
            string
        password (str): Password which must not be blank or contain the :
            character
    """
    # Endpoint
    if endpoint is None:
        raise exceptions.ApiClientConfigurationError(
            'Endpoint URL has not been set, cannot create client'
        )
    elif not isinstance(endpoint, str):
        raise exceptions.ApiClientConfigurationError(
            'Endpoint URL is not a string, cannot create client'
        )
    elif not validate_endpoint(endpoint):
        raise exceptions.ApiClientConfigurationError(
            'Endpoint URL is invalid as an endpoint URI.\n'
            'Endpoints must be https, must not contain usernames, passwords, '
            'param strings, or fragments, and must have either no port or a'
            'valid port.'
        )
    
    # Username
    elif username is None:
        raise exceptions.ApiClientConfigurationError(
            'Username is not set, cannot create client'
        )
    elif not isinstance(username, str):
        raise exceptions.ApiClientConfigurationError(
            'Username is not a string, cannot create client'
        )
    elif ':' in username:
        raise exceptions.ApiClientConfigurationError(
            'Username contains ":" character (illegal in basic auth), '
            'cannot create client'
        )
    
    # Password
    elif password is None:
        raise exceptions.ApiClientConfigurationError(
            'Password is not set, cannot create client'
        )
    elif not isinstance(password, str):
        raise exceptions.ApiClientConfigurationError(
            'Password is not a string, cannot create client'
        )
    elif ':' in password:
        raise exceptions.ApiClientConfigurationError(
            'Password contains ":" character (illegal in basic auth), '
            'cannot create client'
        )
    else:
        return True


@utils.debug_log_call()
def from_env() -> AccessServerManagementClient:
    """Gets parameters for Access Server from environment variables

    The environment variables we are looking for here are:

    * `PYOVPN_AS_ENDPOINT_URL` - The endpoint of the Access Server API
    * `PYOVPN_AS_USERNAME` - Username to authenticate
    * `PYOVPN_AS_PASSWORD` - Password to authenticate

    The below environment variables are optional, and may be set to either
    "true" or "false"
    
    * `PYOVPN_AS_DEBUG` - Are we in debug mode, defaults to false
    * `PYOVPN_AS_ALLOW_UNTRUSTED` - Allow untrusted/unverified SSL context from
        Access Server, defaults to false

    Raises:
        ApiClientConfigurationError: Environment variables were invalid in some 
            way

    Returns:
        AccessServerManagementClient: configured using the above values
    """
    endpoint = os.environ.get('PYOVPN_AS_ENDPOINT_URL')
    username = os.environ.get('PYOVPN_AS_USERNAME')
    password = os.environ.get('PYOVPN_AS_PASSWORD')
    
    validate_client_args(endpoint, username, password)

    debug_str = os.environ.get('PYOVPN_AS_DEBUG', 'false')
    allow_untrusted_str = os.environ.get('PYOVPN_AS_ALLOW_UNTRUSTED', 'false')

    if debug_str.lower() == 'true':
        debug = True
    elif debug_str.lower() == 'false':
        debug = False
    else:
        raise exceptions.ApiClientConfigurationError(
            'PYOVPN_AS_DEBUG must be either true or false'
        )

    if allow_untrusted_str.lower() == 'true':
        allow_untrusted = True
    elif allow_untrusted_str.lower() == 'false':
        allow_untrusted = False
    else:
        raise exceptions.ApiClientConfigurationError(
            'PYOVPN_AS_ALLOW_UNTRUSTED must be either true or false'
        )

    logging.info(
        f'Creating client with ({repr(endpoint)}, {repr(username)}, '
        f'REDACTED, {repr(debug)}, {repr(allow_untrusted)})'
    )
    return AccessServerManagementClient(
        endpoint, username, password, debug=debug,
        allow_untrusted=allow_untrusted
    )


@utils.debug_log_call()
def from_file(filepath: os.PathLike) -> AccessServerManagementClient:
    """Retrieve client configuration from file

    The file must be a JSON file with a single object, specifying the following:
        endpoint_url (str): The endpoint of the AS RPC API (https://<ip>/RPC2)
        username (str): Username to authenticate as
        password (str): Password for the above
        debug (bool, optional): Whether to provide debug information, default
            is false
        allow_untrusted (bool, optional): Whether to allow unverified SSL certs
            from server, default is false, should be turned off in production

    Args:
        filepath (os.PathLike): JSON file for configuration

    Raises:
        ApiClientConfigurationError: JSON configuration was invalid in some way

    Returns:
        AccessServerManagementClient: The client configured using the above 
            options
    """
    with open(filepath) as config_file:
        config = json.loads(config_file.read())
    
    endpoint = config.get('endpoint_url')
    username = config.get('username')
    password = config.get('password')

    validate_client_args(endpoint, username, password)

    debug = config.get('debug', False)
    allow_untrusted = config.get('allow_untrusted', False)

    if not isinstance(debug, bool) or not isinstance(allow_untrusted, bool):
        raise exceptions.ApiClientConfigurationError(
            'debug and allow_untrusted must be either true or false'
        )
    
    logging.info(
        f'Creating client with ({repr(endpoint)}, {repr(username)}, '
        f'REDACTED, {repr(debug)}, {repr(allow_untrusted)})'
    )
    return AccessServerManagementClient(
        endpoint, username, password, debug=debug,
        allow_untrusted=allow_untrusted
    )


@utils.debug_log_call(redact=[2, 'password'])
def from_args(
    endpoint: str,
    username: str,
    password: str,
    debug: bool=False,
    allow_untrusted: bool=False
) -> AccessServerManagementClient:
    """Connect to AccessServer using config specified in arguments

    Args:
        endpoint (str): Endpoint URL of the Access Server
            (https://hostname/RPC2/)
        username (str): Username to authenticate as
        password (str): Password to authenticate with
        debug (bool, optional): Log debug information. Defaults to False.
        allow_untrusted (bool, optional): Allow unverified SSL context and
            untrusted SSL certs from server. Defaults to False.

    Returns:
        AccessServerManagementClient
    """
    validate_client_args(endpoint, username, password)
    if not isinstance(debug, bool) or not isinstance(allow_untrusted, bool):
        raise exceptions.ApiClientConfigurationError(
            'debug and allow_untrusted must be either true or false'
        )
    logging.info(
        f'Creating client with ({repr(endpoint)}, {repr(username)}, '
        f'REDACTED, {repr(debug)}, {repr(allow_untrusted)})'
    )
    return AccessServerManagementClient(
        endpoint, username, password, debug=debug,
        allow_untrusted=allow_untrusted
    )



"""Contains the RPC client that will handle low-level XML-RPC API calls to the
OpenVPN Access Server.
"""
import builtins
import json
import pathlib
import ssl
import urllib.parse
import xmlrpc.client
from typing import Any

from .exceptions import translate_fault

class _SupportedMethod:
    """Code for this is similar to the same code as in xmlrpc.client for
    _Method class.

    https://github.com/python/cpython/blob/374ee449331bc95d18c37f5032aaea1448462e58/Lib/xmlrpc/client.py#L1107-L1116

    Attributes:
        __send (callable): Callable that we are invoking when we call this
            method
        __name (str): Name of the method we are calling
        METHODS (dict): Definition of methods and their parameters (defined
            at the bottom of the file when read from config file)
        PARAM_KEY (str): Key used to get parameters from method definitions
        TYPE_KEY (str): Key used to get param type from param definition
        DEFAULT_KEY (str): Key used to get default value from param definition
        REQUIRED_KEY (str): Key used to get whether a param is required or not
        NULL_KEY (str): Key used to get whether a param can be null or not
        NAME_KEY (str): Key used to get (our) name of the parameter
    
    Args:
        send (callable): Callable that will be __send
        name (str): Name of method which will be __name

    Raises:
        AttributeError: Unsupported method name passed
    """
    METHODS: dict

    PARAM_KEY = 'params'
    TYPE_KEY = 'type'
    DEFAULT_KEY = 'default'
    REQUIRED_KEY = 'required'
    NULL_KEY = 'null'
    NAME_KEY = 'name'

    def __init__(self, send, name):
        self.__send = send
        self.__name = name
        method = self.METHODS.get(self.__name)
        # Raise attribute error if unsupported method
        if method == None:
            raise AttributeError("'RpcClient' object has no "
                f"attribute '{self.__name.split('.')[0]}'")

    @classmethod
    def validate_param(cls, arg: Any, param_def: dict) -> True:
        """Validates that the given argument matches the definition of the param
        Args:
            arg (Any): Argument to be validated
            param_def (dict): Definition of the parameter
        
        Raises:
            TypeError: Invalid argument

        Returns:
            True if argument is valid for the parameter definition given
        """
        # 1. Check if argument is None and parameter is required
        if arg is None and param_def[cls.NULL_KEY]:
            return True

        # 2. Get type of param for comparison
        if '[' in param_def[cls.TYPE_KEY]:
            param_type = getattr(
                builtins,
                param_def[cls.TYPE_KEY].split(
                    '['
                )[0]
            )
            param_item_type = getattr(
                builtins,
                param_def[cls.TYPE_KEY].split(
                    '['
                )[1][:-1]
            )
        else:
            param_type = getattr(builtins, param_def[cls.TYPE_KEY])
            param_item_type = None

        # 3. Check argument is of correct type
        if not isinstance(arg, param_type):
            raise TypeError(f"Expected {param_def[cls.TYPE_KEY]} for arg "
                f"{param_def[cls.NAME_KEY]}, got {type(arg).__name__}")

        if param_item_type is not None \
            and not all(
                isinstance(elem, param_item_type) for elem in arg
            ):
            raise TypeError(f"Expected {param_def[cls.TYPE_KEY]} for arg "
                f"{param_def[cls.NAME_KEY]}, got wrong item type")

        # 4. Check that we are not allowing an empty string if an argument
        # cannot be null
        if param_type in (list, dict, str) \
            and not param_def[cls.NULL_KEY] \
            and len(arg) == 0:
            raise TypeError(f"Expected non-empty {param_type.__name__} for arg "
                f"{param_def[cls.NAME_KEY]}, got empty {type(arg).__name__}")
        
        # Argument is valid for parameter if we haven't raise issue yet
        return True


    def __call__(self, *args, **kwargs) -> Any:
        """Calls self.__send with the arguments provided after validation

        We validate the arguments passed and if they are valid we pass them to
        self.__send

        Returns:
            Result of calling self.__send with *args and **kwargs

        TODO handle errors returned from API
        TODO be aware that an empty dictionary will pass the 'required' test
        """
        method = self.METHODS[self.__name]
        # 1. Validate we have all required arguments
        current_param = 0
        params = method[self.PARAM_KEY]
        # 1a. Positional arguments first
        for arg in args:
            # How many params have we tried to use
            if len(params) == current_param:
                raise TypeError(f"{self.__name} expected at most {len(params)}"
                    f" arguments, got {len(args) + len(kwargs)}")
            param_def = params[current_param]
            self.validate_param(arg, param_def)
            current_param += 1

        params_to_submit = list(args)
        
        # 1b. Keywords next
        param_pool = {
            p[self.NAME_KEY] : {**p, **{'index': i + current_param}} 
            for i, p in enumerate(params[current_param:])
        }
        for kwarg in kwargs:
            if len(params) == current_param:
                raise TypeError(f"{self.__name} expected at most {len(params)}"
                    f" arguments, got {len(args) + len(kwargs)}")
            param_def = param_pool.get(kwarg)
            if param_def is None:
                raise TypeError(f"{self.__name}() got an unexpected keyword "
                    f"argument '{kwarg}'")
            self.validate_param(kwargs[kwarg], param_def)
            current_param += 1

        params_to_submit += [
            kwargs[p[self.NAME_KEY]] for p in param_pool.values()
            if p[self.NAME_KEY] in kwargs
        ]

        # 1c. Add null/default for required params not accounted for 
        # that accept it
        for p in param_pool.values():
            if p[self.REQUIRED_KEY] \
                and (
                    p[self.NULL_KEY]
                    or p.get(self.DEFAULT_KEY)
                ) \
                and p[self.NAME_KEY] not in kwargs:
                params_to_submit.insert(
                    p['index'],
                    p.get(self.DEFAULT_KEY)
                )
                current_param += 1
            elif p[self.REQUIRED_KEY] \
                and not (
                    p[self.NULL_KEY]
                    or p.get(self.DEFAULT_KEY)
                ) \
                and p[self.NAME_KEY] not in kwargs:
                raise TypeError(f"{self.__name} missing argument '"
                    f"{p[self.NAME_KEY]}'")
                
        # 2. Call the function, catching and translating any errors
        try:
            return self.__send(*params_to_submit)
        except xmlrpc.client.Fault as fault:
            new_err = translate_fault(fault)
            if fault == new_err:
                raise fault
            else:
                raise new_err from fault



class RpcClient(object):
    """The client to handle low-level XML-RPC calls to OpenVPN AS
    
    Attributes:
        _serv_proxy (xmlrpc.client.ServerProxy): Manages communication with
            OpenVPN XML-RPC endpoint
        _debug (bool): Whether we are in debug mode, this is insecure
        _allow_unsupported (bool): Whether we allow all functions (no official
            support for this) to be called

    Args:
        endpoint (str): The full URI of the XML-RPC endpoint (usually 
            https://<ip-of-access-server>/RPC2/)
        username (str): Username of an admin user on the remote server
        password (str): Password for the above user
        **kwargs: 
            debug (bool): can be set which will provide debug output
            allow_untrusted (bool): trusts untrusted SSL certificates coming
                from the target server
            allow_unsupported (bool): allow all method calls and now just those
                that are officially supported

    Raises:
        ValueError: The username or password contains an illegal character
    """
    def __init__(self, endpoint, username, password, **kwargs):
        if ':' in password:
            raise ValueError('Password cannot contain ":"')
        elif ':' in username:
            raise ValueError('Username cannot contain ":"')

        self._debug = kwargs.get('debug', False)
        self._allow_unsupported= kwargs.get('allow_unsupported', False)

        safe_username = urllib.parse.quote(username, safe='')
        safe_password = urllib.parse.quote(password, safe='')
        auth_string = f'{safe_username}:{safe_password}'
        parsed_endpoint = urllib.parse.urlparse(endpoint)
        parsed_endpoint = parsed_endpoint._replace(
            netloc=f'{auth_string}@{parsed_endpoint.netloc}'
        )
        new_endpoint = urllib.parse.urlunparse(parsed_endpoint)

        if kwargs.get('allow_untrusted', False):
            # Allows untrusted certificates
            ssl_context = ssl._create_unverified_context()
        else:
            # Full trust only, let ServerProxy do the SSL work
            ssl_context = None
        
        self._serv_proxy = xmlrpc.client.ServerProxy(
            new_endpoint,
            allow_none=True,    # Needs set to allow for sending <nil/>
            verbose=self._debug,
            context=ssl_context
        )
        
        # Try to connect to see if we can reach the server
        try:
            self._serv_proxy.GetASLongVersion()
        except xmlrpc.client.Fault as fault:
            self.close()
            new_err = translate_fault(fault)
            if new_err == fault:
                raise fault
            else:
                raise new_err from fault

    def close(self):
        self.__exit__()

    def __enter__(self):
        return self

    def __exit__(self):
        """Cleanup resources on exit, i.e. close the RPC connection
        """
        self._serv_proxy.__exit__()

    def __getattr__(self, attr: str) -> Any:
        """Magic to return either the raw ServerProxy call or a supported 
           method
        """
        if self._allow_unsupported:
            return getattr(self._serv_proxy, attr)
        else:
            return _SupportedMethod(getattr(self._serv_proxy, attr), attr)


# ---------------------------------
# ---- Setup and configuration ----
# ---------------------------------
SUPPORTED_METHODS: dict
BASE_DIR = pathlib.Path(__file__).resolve().parent
with open(BASE_DIR / 'supported_methods.json') as methods_file:
    SUPPORTED_METHODS = json.loads(methods_file.read())
_SupportedMethod.METHODS = SUPPORTED_METHODS



"""This module provides definitions for high level errors relating to the
   functions provided by the root modules in this package.
   
Errors relating to high level functions in the root modules of the pyovpn_as
package will not have a direct mapping at the API and XML-RPC level. Where the
API may deal with issues such as password complexity, authorisation, and
parameter validation, the exceptions defined here will deal with logical errors.

An example of this may be when we attempt to create a new user, but this user
already exists. The API will let us go ahead and overwrite the existing user's 
properties, but this does not make sense from our perspective, and could cause
substantial and confusing issues.

Similarly, if we attempt to add a user to a group, but that group doesn't exist
(e.g. we made a typo), we would generally want the program to tell us, as
indeed the Web Admin UI does, but the API would not tell us, and would happily
let us set the property without a second thought.
"""


class AccessServerBaseException(Exception):
    """Base exception for all exceptions in this module
    """
    pass


class AccessServerProfileExistsError(AccessServerBaseException):
    """Raised when we try to create a user or group that already exists
    """
    pass


class AccessServerProfileNotFoundError(AccessServerBaseException):
    """Raised when we try to access a user or group that does not exist
    """
    pass


class AccessServerProfileCreateError(AccessServerBaseException):
    """Raised when an error is encountered when creating a new profile
    """
    pass


class AccessServerProfileDeleteError(AccessServerBaseException):
    """Raised when an unknown error is encountered when deleting a profile
    """
    pass


class AccessServerClientExistsError(AccessServerBaseException):
    """Raised when we try to create a client record that already exists
    """
    pass


class AccessServerClientCreateError(AccessServerBaseException):
    """Raised when an error is encountered when creating a new client record
    """
    pass


class PasswordGenerationComplexityTimeout(Exception):
    """Raised when we have been unable to generate a suitably complex password
    """
    pass


class AccessServerConfigError(AccessServerBaseException):
    """Raised when the configuration on the server does not permit a given
       operation to be run
    """
    pass


class AccessServerProfileIntegrityError(AccessServerBaseException):
    """Raised when a profile contains a value that wasn't expected. E.g. when the ``type`` property contains anything other than the defined types.
    """
    pass


class AccessServerPropOverwriteError(AccessServerProfileIntegrityError):
    """Raised when one attempts to overwrite a critical profile property that 
    already has a value on the server (e.g. trying to add a user to a group 
    when they are already a part of another group)
    """
    pass


"""Tests functions in the pyovpn_as.client module
"""
import unittest
import unittest.mock

from pyovpn_as import client
from pyovpn_as.api import exceptions


class TestValidateEndpoint(unittest.TestCase):
    """TestCase for the validate_endpoint function
    """

    def test_valid_endpoint_returns_true(self):
        endpoint = 'https://ip-address/RPC2/'
        self.assertTrue(client.validate_endpoint(endpoint))

    def test_http_scheme_returns_false(self):
        endpoint = 'http://ip-address/'
        self.assertFalse(client.validate_endpoint(endpoint))

    def test_ftp_scheme_returns_false(self):
        endpoint = 'ftp://ip-address/'
        self.assertFalse(client.validate_endpoint(endpoint))

    def test_username_in_netloc_returns_false(self):
        endpoint = 'https://username@ip-address/'
        self.assertFalse(client.validate_endpoint(endpoint))

    def test_passsword_in_netloc_returns_false(self):
        endpoint = 'https://:password@ip-address/'
        self.assertFalse(client.validate_endpoint(endpoint))

    def test_only_userpass_in_netloc_returns_false(self):
        endpoint = 'https://username:password@/'
        self.assertFalse(client.validate_endpoint(endpoint))

    def test_blank_netloc_returns_false(self):
        endpoint = 'https:///RPC2/'
        self.assertFalse(client.validate_endpoint(endpoint))

    def test_get_query_returns_false(self):
        endpoint = 'https://ip-address/?param=val'
        self.assertFalse(client.validate_endpoint(endpoint))

    def test_parameters_returns_false(self):
        endpoint = 'https://ip-address/RPC2;param=val'
        self.assertFalse(client.validate_endpoint(endpoint))

    def test_fragments_returns_false(self):
        endpoint = 'https://ip-address/#fragment'
        self.assertFalse(client.validate_endpoint(endpoint))

    def test_valid_port_number_returns_true(self):
        endpoint = 'https://ip-address:443/'
        self.assertTrue(client.validate_endpoint(endpoint))

    def test_high_port_number_returns_false(self):
        endpoint = 'https://ip-address:65536/'
        self.assertFalse(client.validate_endpoint(endpoint))

    def test_negative_port_number_returns_false(self):
        endpoint = 'https://ip-address:-1/'
        self.assertFalse(client.validate_endpoint(endpoint))


@unittest.mock.patch('pyovpn_as.client.validate_endpoint', return_value=True)
class TestValidateClientArgs(unittest.TestCase):
    """TestCase for the validate_client_args function
    """

    def test_valid_args_returns_true(self, *args):
        self.assertTrue(
            client.validate_client_args(
                'https://endpoint/',
                'username',
                'password'
            )
        )

    def test_none_endpoint_raises_error(self, *args):
        with self.assertRaisesRegex(
            exceptions.ApiClientConfigurationError,
            'Endpoint URL has not been set, cannot create client'
        ):
            client.validate_client_args(
                None,
                'username',
                'password'
            )

    def test_non_str_endpoint_raises_error(self, *args):
        with self.assertRaisesRegex(
            exceptions.ApiClientConfigurationError,
            'Endpoint URL is not a string, cannot create client'
        ):
            client.validate_client_args(
                [],
                'username',
                'password'
            )

    def test_none_username_raises_error(self, *args):
        with self.assertRaisesRegex(
            exceptions.ApiClientConfigurationError,
            'Username is not set, cannot create client'
        ):
            client.validate_client_args(
                'https://endpoint/',
                None,
                'password'
            )

    def test_non_str_username_raises_error(self, *args):
        with self.assertRaisesRegex(
            exceptions.ApiClientConfigurationError,
            'Username is not a string, cannot create client'
        ):
            client.validate_client_args(
                'https://endpoint/',
                [],
                'password'
            )

    def test_colon_in_username_raises_error(self, *args):
        with self.assertRaisesRegex(
            exceptions.ApiClientConfigurationError,
            'Username contains ":" character \(illegal in basic auth\), '
            'cannot create client'
        ):
            client.validate_client_args(
                'https://endpoint/',
                'user:name',
                'password'
            )

    def test_none_password_raises_error(self, *args):
        with self.assertRaisesRegex(
            exceptions.ApiClientConfigurationError,
            'Password is not set, cannot create client'
        ):
            client.validate_client_args(
                'https://endpoint/',
                'username',
                None
            )

    def test_non_str_password_raises_error(self, *args):
        with self.assertRaisesRegex(
            exceptions.ApiClientConfigurationError,
            'Password is not a string, cannot create client'
        ):
            client.validate_client_args(
                'https://endpoint/',
                'username',
                []
            )

    def test_colon_in_password_raises_error(self, *args):
        with self.assertRaisesRegex(
            exceptions.ApiClientConfigurationError,
            'Password contains ":" character \(illegal in basic auth\), '
            'cannot create client'
        ):
            client.validate_client_args(
                'https://endpoint/',
                'username',
                'pass:word'
            )

if __name__ == '__main__':
    unittest.main()


import ipaddress
from datetime import datetime
from typing import Any

from pyovpn_as.api import cli


class ClientStatus:
    """Represents the status of a given VPN client connected to the server

    Attributes:
        connected_since (datetime): When the client connected to the server
        username (str): Username of the connected client
        common_name (str): Common name of the certificate used to connect
        bytes_received (int): The bytes received from the client since 
            connecting
        bytes_send (int): The bytes send to the client since connecting
        virtual_address (IPv4Address): The virtual IPv4 address of the 
            connected client
        virtual_ipv6_address (IPv6Address): The virtual IPv6 address, or None 
            if the client has not been assigned one.
        real_address (str): The source IP address and port of the client 
            connection
        client_id (int): ID of the client connection
        peer_id (int): TODO define
    
    Args:
        attributes (list[str]): The list of attributes associated with a client 
            connections
        headers (dict[str, int]): A dictionary mapping header names to index in 
            the attributes list
    """
    def __init__(self, attributes: list, headers: dict):
        self.connected_since = datetime.fromtimestamp(attributes[
            headers['Connected Since (time_t)']
        ])
        self.username = attributes[
            headers['Username']
        ]
        self.common_name = attributes[
            headers['Common Name']
        ]
        self.bytes_received = int(attributes[
            headers['Bytes Received']
        ])
        self.bytes_sent = int(attributes[
            headers['Bytes Sent']
        ])
        self.virtual_address = ipaddress.ip_address(
            attributes[headers['Virtual Address']]
        )
        vipv6 = attributes[headers['Virtual IPv6 Address']]
        if vipv6 == '':
            self.virtual_ipv6_address = None
        else:
            self.virtual_ipv6_address = ipaddress.ip_address(vipv6)
        self.real_address = attributes[
            headers['Real Address']
        ]
        self.client_id = int(attributes[
            headers['Client ID']
        ])
        self.peer_id = int(attributes[
            headers['Peer ID']
        ])


class VPNStatus:
    """Represents the status of a given VPN interface

    We can get these connections from the sacli.VPNStatus() method. We access 
    the client_list key/value pair of the openvpn_X dictionaries and gather the 
    statistics from these dictionaries.

    Each openvpn_X dictionary will have a dictionary defining the headers for 
    the client_list lists and will look something like this::

        "client_list_header": {
            "Connected Since": 6,
            "Username": 8,
            "Common Name": 0,
            "Bytes Received": 4,
            "Virtual Address": 2,
            "Bytes Sent": 5,
            "Virtual IPv6 Address": 3,
            "Client ID": 9,
            "Connected Since (time_t)": 7,
            "Real Address": 1,
            "Peer ID": 10
        }
    
    The above indicates that the Common Name (of the certificate used to 
    connect) is in position 0 of each of the below lists::

        "client_list": [
            [
                "Example_Username",
                "1.1.1.1:55555",
                "172.27.228.2",
                "",
                "143313",
                "2727656",
                "Tue May  4 13:54:03 2021",
                "1620136443",
                "Example_Username",
                "0",
                "0"
            ]
        ]

    Attributes:
        interface_name (str): The name of the VPN daemon this status refers to
        connected_clients (list[ClientStatus]): List of clients connected to 
            the VPN

    Args:
        daemon_name (str): The name of the daemon the status refers to. 
            Normally this will be a name of the form ``openvpn_X``
        connection_summary (dict[str, Any]): A dictionary representing the 
            status of connections of a given VPN interface

    Raises:
        TypeError: if the connection_summary is not a dictionary
    """
    def __init__(self, daemon_name: str, connection_summary: dict):
        if not isinstance(connection_summary, dict):
            raise TypeError(
                f"Expected 'dict' for arg 'connection_summary', got "
                f"'{type(connection_summary)}'"
            )
        if not isinstance(daemon_name, str):
            raise TypeError(
                f"Expected 'str' for arg 'daemon_name', got "
                f"'{type(daemon_name)}'"
            )
        
        self.daemon_name = daemon_name

        headers = connection_summary['client_list_headers']
        self.connected_clients = [
            ClientStatus(attrs, headers) 
            for attrs in connection_summary['client_list']
        ]


class VpnOperations:
    """This class provides methods to monitor the status and alter the configuration of the VPN Daemon service

    Args:
        sacli (RemoteSacli): The client used to communicate with the server
    """
    def __init__(self, sacli: cli.RemoteSacli):
        if not isinstance(sacli, cli.RemoteSacli):
            raise TypeError(
                f"Expected 'RemoteSacli' for arg 'sacli', got '{type(sacli)}'"
            )
        self._sacli = sacli

    
    @property
    def status(self) -> list:
        """list[VpnStatus]: The detailed status of connections to the VPN 
        daemons
        """
        status_dict = self._sacli.GetVpnStatus()
        return [
            VPNStatus(daemon, status)
            for daemon, status 
            in status_dict
        ]



"""Tests the classes in pyovpn_as.api.cli
"""
import unittest

from pyovpn_as.api import cli
from pyovpn_as.api.exceptions import ApiClientPasswordComplexityError

class TestIsPasswordComplex(unittest.TestCase):
    """This TestCase tests the is_password_complex static method of the
       AccessServerClient class
    """

    def test_compliant_password_returns_true(self):
        password = 'Th1sIs4C0mpliantPassw0rd%'
        try:
            self.assertTrue(
                cli.RemoteSacli.is_password_complex(password)
            )
        except Exception as err:
            self.fail(f'Password complexity check failed: {err}')

    def test_short_password_raises_error(self):
        password = '$T2a'
        with self.assertRaises(ApiClientPasswordComplexityError):
            cli.RemoteSacli.is_password_complex(password)

    def test_7_char_password_raises_error(self):
        password = '$T2a123'
        with self.assertRaises(ApiClientPasswordComplexityError):
            cli.RemoteSacli.is_password_complex(password)

    def test_8_char_password_returns_true(self):
        password = '$T2a1234'
        try:
            self.assertTrue(
                cli.RemoteSacli.is_password_complex(password)
            )
        except Exception as err:
            self.fail(f'Password complexity check failed: {err}')

    def test_very_long_password_returns_true(self):
        password = f'$T2a{"a" * 1000000}'
        try:
            self.assertTrue(
                cli.RemoteSacli.is_password_complex(password)
            )
        except Exception as err:
            self.fail(f'Password complexity check failed: {err}')

    def test_none_password_raises_error(self):
        password = None
        with self.assertRaises(ApiClientPasswordComplexityError):
            cli.RemoteSacli.is_password_complex(password)

    def test_non_str_password_raises_TypeError(self):
        password = []
        with self.assertRaises(TypeError):
            cli.RemoteSacli.is_password_complex(password)

    def test_password_no_digit_raises_error(self):
        password = 'abcdE$Gh'
        with self.assertRaises(ApiClientPasswordComplexityError):
            cli.RemoteSacli.is_password_complex(password)

    def test_password_no_lowercase_raises_error(self):
        password = 'ABCDE$G1'
        with self.assertRaises(ApiClientPasswordComplexityError):
            cli.RemoteSacli.is_password_complex(password)

    def test_password_no_uppercase_raises_error(self):
        password = 'abcde$g1'
        with self.assertRaises(ApiClientPasswordComplexityError):
            cli.RemoteSacli.is_password_complex(password)

    def test_password_with_each_symbol_returns_true(self):
        password = 'Th1sIs4C0mpliantPassw0rd%'
        for sym in "!@#$%&'()*+,-/[\\]^_`{|}~<>.":
            password = password[:-1] + sym
            try:
                self.assertTrue(
                    cli.RemoteSacli.is_password_complex(password)
                )
            except ApiClientPasswordComplexityError as err:
                self.fail(
                    f'Password complexity check failed with symbol {sym}: {err}'
                )

    def test_password_with_disallowed_symbols_raises_error(self):
        password = 'abcde:g1'
        with self.assertRaises(ApiClientPasswordComplexityError):
            cli.RemoteSacli.is_password_complex(password)

    def test_password_with_control_characters_raises_error(self):
        password = 'abcde\x08g1'
        with self.assertRaises(ApiClientPasswordComplexityError):
            cli.RemoteSacli.is_password_complex(password)

    def test_password_with_whitespace_raises_error(self):
        password = 'abcde\ng1'
        with self.assertRaises(ApiClientPasswordComplexityError):
            cli.RemoteSacli.is_password_complex(password)

    def test_password_with_non_ascii_raises_error(self):
        password = 'abcde\u0394g1'
        with self.assertRaises(ApiClientPasswordComplexityError):
            cli.RemoteSacli.is_password_complex(password)


if __name__ == '__main__':
    unittest.main()


"""A module containing possible exceptions that may arise during communication
   with the XML-RPC endpoint for Access Server
"""
import xmlrpc.client

class ApiClientBaseException(Exception):
    """Used as a base exception for other errors in the API conversation
       that are not XMLRPC Faults, e.g. bad password in SetLocalPassword
    """
    pass

class ApiClientAuthError(ApiClientBaseException):
    """Given by Fault Code 9007, this error pertains to the user either not 
       having correct permissions or their login being incorrect.
    """
    pass

class ApiClientParameterError(ApiClientBaseException):
    """Given by Fault Code 8002, this error indicates that the number of 
       parameters passed to the given method is incorrect
    """
    pass

class ApiClientValueError(ApiClientBaseException):
    """The Fault Code for this exception will be 9000, but the Fault String
       will begin with something like 'XMLRPCRelay: exceptions.ValueError:'
    """
    pass

class ApiClientInternalError(ApiClientBaseException):
    """Raised when the Fault Code is 9000 and Fault String is "XMLRPC: 
       internal error"
    """
    pass

class ApiClientFunctionNotFoundError(ApiClientBaseException):
    """Raised when the Fault Code is 9000 and Fault String is "XMLRPCRelay:
       XMLRPC: function not found"
    """
    pass

class ApiClientPasswordComplexityError(ApiClientBaseException):
    """Raised when the new password sent to the server during a password change
       does not meet the complexity requirements set by the server.

    Generally speaking, the Access Server requires passwords to be at least 8
    characters long, contain an uppercase letter, a lowercase letter, a digit
    and a symbol from !@#$%&'()*+,-/[\\]^_`{|}~<>. (full stop included, also
    note the absence of colon and double quotation marks).
    """
    pass

class ApiClientPasswordIncorrectError(ApiCientAuthError):
    """Raised when the current password sent to the server during a password
       password change method call (e.g. SetLocalPassword) is incorrect
    """
    pass

class ApiClientPasswordResetError(ApiClientBaseException):
    """Raised when something goes wrong that was not expected during a password
       change method call (e.g. SetLocalPassword)
    """
    pass

class ApiClientUnexpectedError(ApiClientBaseException):
    """Raised when an error occurs that we hadn't accounted for. If this is
       raised we should review the error closely and create a new exception
       class for it
    """
    pass

class ApiClientConfigurationError(ApiClientBaseException):
    """Raised when the configuration we have attempted to use for creating an
       AccessServerClient is invalid, e.g. no endpoint configured
    """
    pass

class ApiClientUsernameNotFoundError(ApiClientBaseException):
    """Raised when the Fault Code is 9000 and the Fault String is
       "XMLRPCRelay: AUTHRPC_EXCEPT: CertDB: Username 'x' not found"
    """
    pass

# -----------------
# ---- Methods ----
# -----------------
def translate_fault(err: Exception) -> None:
    """Translates a given exception into one more friendly for the user

    We return the exception rather than raise it because no error has
    occurred in this method. We return so that it can be raised in the right
    place to prevent confusion.

    Args:
        err (Exception): The Exception to translate

    Returns:
        Exception: this time it's the correct type with message and all
    """
    if not isinstance(err, xmlrpc.client.Fault):
        return err
    elif err.faultCode == 8002:
        return ApiClientParameterError(
            'Number of parameters is incorrect'
        )
    elif err.faultCode == 9007:
        return ApiCientAuthError(
            'Either your credentials are wrong or your permissions are not'
            ' correct to run the given method.'
        )
    elif err.faultCode == 9000 and err.faultString.startswith(
        'XMLRPCRelay: exceptions.ValueError: '
    ):
        start_from = len('XMLRPCRelay: exceptions.ValueError: ')
        return ApiClientValueError(
            f'ValueError from server: {err.faultString[start_from:]}'
        )
    elif err.faultCode == 9000 and \
        err.faultString == 'XMLRPC: internal error':
        return ApiClientInternalError(
            'Something unknown went wrong with that call that the server '
            'did not like'
        )
    elif err.faultCode == 9000 and \
        err.faultString == 'XMLRPCRelay: XMLRPC: function not found':
        return ApiClientFunctionNotFoundError(
            'Function not found on given server'
        )
    elif err.faultCode == 9000 and \
        err.faultString.startswith(
            'XMLRPCRelay: AUTHRPC_EXCEPT: CertDB: Username'
        ) and err.faultString.endswith(
            'not found'
        ):
        return ApiClientUsernameNotFoundError(
            'Username could not be found on the server'
        )
    else:
        return ApiClientUnexpectedError(
            'Something happened that we were not expecting.\n'
            f'Fault Code: {err.faultCode}\n'
            f'Fault String: "{err.faultString}"'
        )





"""Tests the classes in pyovpn_as.api.rpc 
"""
import unittest
import unittest.mock

from pyovpn_as.api import rpc

@unittest.mock.patch.dict(
    rpc._SupportedMethod.METHODS, {
        'method_name': 'test'
    }, clear=True
)
class TestSupportedMethodInit(unittest.TestCase):
    """Tests the functionality of the __init__ method in the _SupportedMethod
       class
    """
    KNOWN_METHOD = 'method_name'

    def test_supported_method_init_wrong_name_raises_AttributeError(self):
        """Tests that an AttributeError is raised when a method name is given
           that is not supported.
        """
        with self.assertRaisesRegex(
            AttributeError,
            "object has no attribute 'not_a_name'"
        ):
            rpc._SupportedMethod(None, 'not_a_name')
    
    def test_supported_method_init_right_name_raises_nothing(self):
        """Tests that an object is created successfully when a method name is
           supported
        """
        self.assertIsInstance(
            rpc._SupportedMethod(None, self.KNOWN_METHOD),
            rpc._SupportedMethod
        )


@unittest.mock.patch.dict('pyovpn_as.api.rpc._SupportedMethod.METHODS', {
    "TestMethod": {
        rpc._SupportedMethod.PARAM_KEY: [
            {
                rpc._SupportedMethod.NAME_KEY: "bool",
                rpc._SupportedMethod.TYPE_KEY: "bool",
                rpc._SupportedMethod.NULL_KEY: False
            },
            {
                rpc._SupportedMethod.NAME_KEY: "int",
                rpc._SupportedMethod.TYPE_KEY: "int",
                rpc._SupportedMethod.NULL_KEY: False
            },
            {
                rpc._SupportedMethod.NAME_KEY: "float",
                rpc._SupportedMethod.TYPE_KEY: "float",
                rpc._SupportedMethod.NULL_KEY: False
            },
            {
                rpc._SupportedMethod.NAME_KEY: "str",
                rpc._SupportedMethod.TYPE_KEY: "str",
                rpc._SupportedMethod.NULL_KEY: False
            },
            {
                rpc._SupportedMethod.NAME_KEY: "str - null",
                rpc._SupportedMethod.TYPE_KEY: "str",
                rpc._SupportedMethod.NULL_KEY: True
            },
            {
                rpc._SupportedMethod.NAME_KEY: "list[str]",
                rpc._SupportedMethod.TYPE_KEY: "list[str]",
                rpc._SupportedMethod.NULL_KEY: False
            },
            {
                rpc._SupportedMethod.NAME_KEY: "list[str] - null",
                rpc._SupportedMethod.TYPE_KEY: "list[str]",
                rpc._SupportedMethod.NULL_KEY: True
            },
            {
                rpc._SupportedMethod.NAME_KEY: "dict[str]",
                rpc._SupportedMethod.TYPE_KEY: "dict[str]",
                rpc._SupportedMethod.NULL_KEY: False
            },
            {
                rpc._SupportedMethod.NAME_KEY: "dict[str] - null",
                rpc._SupportedMethod.TYPE_KEY: "dict[str]",
                rpc._SupportedMethod.NULL_KEY: True
            },
        ] 
    }
}, clear=True)
class TestSupportedMethodValidateParam(unittest.TestCase):
    """Tests the functionality of rpc._SupportedMethod.validate_param
    """
    # ------------------------------
    # ---- bool 
    # ------------------------------
    def test_bool_param_bool_arg_passes_validation(self):
        param = rpc._SupportedMethod.METHODS['TestMethod']['params'][0]
        self.assertTrue(
            rpc._SupportedMethod.validate_param(True, param)
        )
    
    def test_bool_param_non_bool_arg_raises_TypeError(self):
        param = rpc._SupportedMethod.METHODS['TestMethod']['params'][0]
        with self.assertRaisesRegex(
            TypeError, 'Expected bool for arg bool, got int'
        ):
            rpc._SupportedMethod.validate_param(3, param)
    
    # ------------------------------
    # ---- int 
    # ------------------------------
    def test_int_param_int_arg_passes_validation(self):
        param = rpc._SupportedMethod.METHODS['TestMethod']['params'][1]
        self.assertTrue(
            rpc._SupportedMethod.validate_param(3, param)
        )
    
    def test_int_param_non_int_arg_raises_TypeError(self):
        param = rpc._SupportedMethod.METHODS['TestMethod']['params'][1]
        with self.assertRaisesRegex(
            TypeError, 'Expected int for arg int, got str'
        ):
            rpc._SupportedMethod.validate_param('test', param)
    
    # ------------------------------
    # ---- float 
    # ------------------------------
    def test_float_param_float_arg_passes_validation(self):
        param = rpc._SupportedMethod.METHODS['TestMethod']['params'][2]
        self.assertTrue(
            rpc._SupportedMethod.validate_param(3.2, param)
        )
    
    def test_float_param_non_float_arg_raises_TypeError(self):
        param = rpc._SupportedMethod.METHODS['TestMethod']['params'][2]
        with self.assertRaisesRegex(
            TypeError, 'Expected float for arg float, got str'
        ):
            rpc._SupportedMethod.validate_param('test', param)
    
    # ------------------------------
    # ---- str 
    # ------------------------------
    def test_str_param_str_arg_passes_validation(self):
        param = rpc._SupportedMethod.METHODS['TestMethod']['params'][3]
        self.assertTrue(
            rpc._SupportedMethod.validate_param('test', param)
        )
    
    def test_str_param_non_str_arg_raises_TypeError(self):
        param = rpc._SupportedMethod.METHODS['TestMethod']['params'][3]
        with self.assertRaisesRegex(
            TypeError, 'Expected str for arg str, got int'
        ):
            rpc._SupportedMethod.validate_param(3, param)

    def test_required_param_none_arg_raises_TypeError(self):
        param = rpc._SupportedMethod.METHODS['TestMethod']['params'][3]
        with self.assertRaisesRegex(
            TypeError, 'Expected str for arg str, got NoneType'
        ):
            rpc._SupportedMethod.validate_param(None, param)

    def test_required_str_param_none_arg_raises_TypeError(self):
        param = rpc._SupportedMethod.METHODS['TestMethod']['params'][3]
        with self.assertRaisesRegex(
            TypeError, 'Expected non-empty str for arg str, got empty str'
        ):
            rpc._SupportedMethod.validate_param('', param)

    def test_optional_param_none_arg_passes_validation(self):
        param = rpc._SupportedMethod.METHODS['TestMethod']['params'][4]
        self.assertTrue(
            rpc._SupportedMethod.validate_param(None, param)
        )

    def test_optional_param_empty_str_arg_passes_validation(self):
        param = rpc._SupportedMethod.METHODS['TestMethod']['params'][4]
        self.assertTrue(
            rpc._SupportedMethod.validate_param('', param)
        )
    
    # ------------------------------
    # ---- list 
    # ------------------------------
    def test_list_param_list_arg_passes_validation(self):
        param = rpc._SupportedMethod.METHODS['TestMethod']['params'][5]
        self.assertTrue(
            rpc._SupportedMethod.validate_param(['test',], param)
        )
    
    def test_list_param_non_list_arg_raises_TypeError(self):
        param = rpc._SupportedMethod.METHODS['TestMethod']['params'][5]
        with self.assertRaisesRegex(
            TypeError, 'Expected list\[str] for arg list\[str], got str'
        ):
            rpc._SupportedMethod.validate_param('test', param)
    
    def test_list_str_param_non_list_str_arg_raises_TypeError(self):
        param = rpc._SupportedMethod.METHODS['TestMethod']['params'][5]
        with self.assertRaisesRegex(
            TypeError,
            'Expected list\[str] for arg list\[str], got wrong item type'
        ):
            rpc._SupportedMethod.validate_param(['test', 3], param)
    
    def test_required_list_param_empty_list_arg_raises_TypeError(self):
        param = rpc._SupportedMethod.METHODS['TestMethod']['params'][5]
        with self.assertRaisesRegex(
            TypeError,
            'Expected non-empty list for arg list\[str], got empty list'
        ):
            rpc._SupportedMethod.validate_param([], param)

    def test_optional_list_param_empty_list_arg_passes_validation(self):
        param = rpc._SupportedMethod.METHODS['TestMethod']['params'][6]
        self.assertTrue(
            rpc._SupportedMethod.validate_param([], param)
        )
    
    # ------------------------------
    # ---- dict 
    # ------------------------------
    def test_dict_param_dict_arg_passes_validation(self):
        param = rpc._SupportedMethod.METHODS['TestMethod']['params'][7]
        self.assertTrue(
            rpc._SupportedMethod.validate_param({'test': 'test'}, param)
        )
    
    def test_dict_param_non_dict_arg_raises_TypeError(self):
        param = rpc._SupportedMethod.METHODS['TestMethod']['params'][7]
        with self.assertRaisesRegex(
            TypeError, 'Expected dict\[str] for arg dict\[str], got str'
        ):
            rpc._SupportedMethod.validate_param('test', param)
    
    def test_dict_str_param_non_dict_str_arg_raises_TypeError(self):
        param = rpc._SupportedMethod.METHODS['TestMethod']['params'][7]
        with self.assertRaisesRegex(
            TypeError,
            'Expected dict\[str] for arg dict\[str], got wrong item type'
        ):
            rpc._SupportedMethod.validate_param({'test': 'test', 3: 3}, param)
    
    def test_required_dict_param_empty_dict_arg_raises_TypeError(self):
        param = rpc._SupportedMethod.METHODS['TestMethod']['params'][7]
        with self.assertRaisesRegex(
            TypeError,
            'Expected non-empty dict for arg dict\[str], got empty dict'
        ):
            rpc._SupportedMethod.validate_param({}, param)

    def test_optional_dict_param_empty_dict_arg_passes_validation(self):
        param = rpc._SupportedMethod.METHODS['TestMethod']['params'][8]
        self.assertTrue(
            rpc._SupportedMethod.validate_param({}, param)
        )

test_method_def = {
    "TestMethod": {
        rpc._SupportedMethod.PARAM_KEY: [
            {
                rpc._SupportedMethod.NAME_KEY: "required",
                rpc._SupportedMethod.REQUIRED_KEY: True,
                rpc._SupportedMethod.NULL_KEY: False,
            },
            {
                rpc._SupportedMethod.NAME_KEY: "required_null",
                rpc._SupportedMethod.REQUIRED_KEY: True,
                rpc._SupportedMethod.NULL_KEY: True,
            },
            {
                rpc._SupportedMethod.NAME_KEY: "required_default",
                rpc._SupportedMethod.REQUIRED_KEY: True,
                rpc._SupportedMethod.NULL_KEY: False,
                rpc._SupportedMethod.DEFAULT_KEY: "default",
            },
            {
                rpc._SupportedMethod.NAME_KEY: "required_null_default",
                rpc._SupportedMethod.REQUIRED_KEY: True,
                rpc._SupportedMethod.NULL_KEY: True,
                rpc._SupportedMethod.DEFAULT_KEY: "default",
            },
            {
                rpc._SupportedMethod.NAME_KEY: "not_required",
                rpc._SupportedMethod.REQUIRED_KEY: False,
                rpc._SupportedMethod.NULL_KEY: False,
            },
            {
                rpc._SupportedMethod.NAME_KEY: "not_required_2",
                rpc._SupportedMethod.REQUIRED_KEY: False,
                rpc._SupportedMethod.NULL_KEY: False,
            }
        ]
    }
}

@unittest.mock.patch.dict(
    'pyovpn_as.api.rpc._SupportedMethod.METHODS',
    test_method_def,
    clear=True
)
@unittest.mock.patch.object(
    rpc._SupportedMethod,
    'validate_param',
    unittest.mock.Mock(return_value=True)
)
class TestSupportedMethodCall(unittest.TestCase):
    """Tests the functionality of rpc._SupportedMethod.__call__
    """
    @unittest.mock.patch.dict(
        'pyovpn_as.api.rpc._SupportedMethod.METHODS',
        test_method_def,
        clear=True
    )
    def setUp(self):
        self.send_mock = unittest.mock.Mock(return_value='test')
        self.method = rpc._SupportedMethod(self.send_mock, 'TestMethod')
    
    def test_more_args_than_params_raises_TypeError(self):
        with self.assertRaisesRegex(
            TypeError,
            "TestMethod expected at most [0-9]+ arguments, got [0-9]+"
        ):
            self.method(1, 2, 3, 4, 5, 6, 7)

    def test_more_kwargs_than_params_raises_TypeError(self):
        with self.assertRaisesRegex(
            TypeError,
            "TestMethod expected at most [0-9]+ arguments, got [0-9]+"
        ):
            self.method(
                required=1,
                required_null=2,
                required_default=3,
                required_null_default=4,
                not_required=5,
                not_required_2=6,
                doesnt_exist=7
            )
    
    def test_more_args_kwargs_than_params_raises_TypeError(self):
        with self.assertRaisesRegex(
            TypeError,
            "TestMethod expected at most [0-9]+ arguments, got [0-9]+"
        ):
            self.method(
                1, 2, 3,
                required_null_default=4,
                not_required=5,
                not_required_2=6,
                doesnt_exist=7
            )

    def test_duplicate_arg_kwarg_raises_TypeError(self):
        with self.assertRaisesRegex(
            TypeError,
            "TestMethod\(\) got multiple values for argument 'required'"
        ):
            self.method(
                1, 2, 3, 4,
                required='duplicate value'
            )
    
    def test_non_existent_kwarg_raises_TypeError(self):
        with self.assertRaisesRegex(
            TypeError,
            "TestMethod\(\) got an unexpected keyword argument 'unexpected'"
        ):
            self.method(
                1, 2, 3, 4,
                unexpected='wow, this is unexpected'
            )
    
    def test_required_no_default_not_null_raises_TypeError(self):
        with self.assertRaisesRegex(
            TypeError,
            "TestMethod missing argument 'required'"
        ):
            self.method(
                required_null=2,
                required_default=3,
                required_null_default=4,
                not_required=5
            )
    
    def test_required_no_default_null_calls_with_null(self):
        self.method(
            required=1,
            required_default=3,
            required_null_default=4,
            not_required=5
        )
        self.send_mock.assert_called_with(1, None, 3, 4, 5)
    
    def test_required_default_no_null_calls_with_default(self):
        self.method(
            required=1,
            required_null=2,
            required_null_default=4,
            not_required=5
        )
        self.send_mock.assert_called_with(1, 2, 'default', 4, 5)
    
    def test_required_default_null_calls_with_default(self):
        self.method(
            required=1,
            required_null=2,
            required_default=3,
            not_required=5
        )
        self.send_mock.assert_called_with(1, 2, 3, 'default', 5)

    def test_skip_not_required_parameter_raises_TypeError(self):
        with self.assertRaisesRegex(
            TypeError,
            "TestMethod missing argument 'not_required'"
        ):
            self.method(
                required=1,
                required_null=2,
                required_default=3,
                required_null_default=4,
                not_required_2=6
            )



if __name__ == '__main__':
    unittest.main()
        


"""This module provides functionality similar to that of the sacli tool in
   OpenVPN Access Servers.

Attributes:
    XML_RPC_VAL (TypeVar): The list of types that an XML-RPC parameter is
        allowed to be
"""

import logging
from datetime import datetime
from typing import TypeVar

from pyovpn_as.api.exceptions import (ApiClientParameterError,
                                      ApiClientPasswordComplexityError,
                                      ApiClientPasswordIncorrectError,
                                      ApiClientPasswordResetError)

from .rpc import RpcClient

logger = logging.getLogger(__name__)

XML_RPC_VAL = TypeVar(
    'XML_RPC_VAL',
    str,
    bool,
    int,
    float,
    dict,
    list,
    tuple,
    type(None),
    bytes,
    bytearray,
    datetime
)

class RemoteSacli:
    """A Wrapper around RpcClient that allows us to emulate the functionality of
       the sacli tool provided in OpenVPN Access Server installations.

    This class is essentially a wrapper around the RpcClient which will handle
    multiple XML-RPC calls if needed for a given operation. This class also
    allows us to handle more complex errors returned by the RpcClient.

    Attributes:
        _RpcClient (RpcClient): Client object that we are wrapping

    Args:
        endpoint (str): The XML-RPC endpoint to connect to, usually in the form
            https://<ip-address>/RPC2/
        username (str): Username of an admin user on the server to connect as
        password (str): Password of the above user that we can connect using
        debug (bool, optional): Provide verbose debug output from the RpcClient.
            Defaults to False
        allow_untrusted (bool, optional): Allow endpoint to send SSL cert that
            is unverified or untrusted. This should ideally be turned off in
            production to prevent MITM attacks. Defaults to False.
    
    Raises:
        ValueError: Password contains and illegal character
        ApiClientBaseException: Something went wrong in connecting to the
            AccessServer endpoint
    """
    def __init__(
        self,
        endpoint: str,
        username: str,
        password: str,
        debug=False,
        allow_untrusted=False
    ):
        self._RpcClient = RpcClient(
            endpoint,
            username,
            password, 
            debug=debug,
            allow_untrusted=allow_untrusted
        )

    def __enter__(self):
        return self

    def __exit__(self):
        self.close()
    
    def close(self):
        self._RpcClient.close()


    @staticmethod
    def is_password_complex(new_pass: str) -> bool:
        """Validate that a password is suitably complex for OpenVPN AS

        An OpenVPN Access Server password must be at least 8 characters long and
        contain a digit, an uppercase letter, a lowercase letter and a symbol
        from !@#$%&'()*+,-/[\\]^_`{|}~<>. (full stop included, also note
        absence of colon and double quotation marks).

        Args:
            new_pass (str): Password to check validate

        Raises:
            ApiClientPasswordComplexityError: Password is not complex enough

        Returns:
            bool: True if the password is suitably complex
        """
        complexity_err = ApiClientPasswordComplexityError(
            "New Password must be at least 8 characters. Password must "
            "also contain a digit, an Uppercase letter, and a symbol from "
            "!@#$%&'()*+,-/[\\]^_`{|}~<>."
        )
        # None
        if new_pass is None:
            raise complexity_err
        # Catch someone not passing string
        if not isinstance(new_pass, str):
            raise TypeError(
                'is_password_complex expected new_pass to be str, got '
                f'{type(new_pass)}'
            )
        # Length
        if len(new_pass) < 8:
            raise complexity_err
        # Uppercase lowercase
        if new_pass.upper() == new_pass or new_pass.lower() == new_pass:
            raise complexity_err
        # Digit
        contains_digit = False
        for i in '0123456789':
            if i in new_pass:
                contains_digit = True
                break
        # Symbol
        contains_symbol = False
        for s in "!@#$%&'()*+,-/[\\]^_`{|}~<>.":
            if s in new_pass:
                contains_symbol = True
                break
        if not contains_symbol or not contains_digit:
            raise complexity_err
        
        return True
            

    def UserPropPut(
        self,
        user: str,
        key: str,
        value: XML_RPC_VAL,
        noui: bool=False
    ) -> list:
        """Add a property to a user profile (create user profile if it doesn't
           exist)

        Args:
            user (str): Username of profile to change
            key (str): Property name to put
            value (XML_RPC_VAL): Value of the property
            noui (bool, optional): Hide user profile in the WebUI. Defaults
                to False.

        Returns:
            list[bool, dict]
        """
        self._RpcClient.UserPropPut(user, {key: value}, noui)
        user_profile = self._RpcClient.UserPropProfileMultiGet(pfilt=[user,])
        return self._RpcClient.UserPropReplace(
            user,
            user_profile[user]
        )

    def UserPropGet(
        self,
        pfilt: list=None,
        tfilt: list=None
    ) -> dict:
        """Retrieves a list of profiles from the server, filtering on profile
           name and profile type (e.g. user_connect or user_connect_hidden)

        Args:
            tfilt (list[str], optional): List of profile types to filter by.
                Defaults to None.
            pfilt (list[str], optional): List of profile names to filter by.
                Defaults to None.

        Returns:
            dict[str, dict[str, XML_RPC_VAL]]: Dictionary of profiles retrieved
        """
        return self._RpcClient.UserPropProfileMultiGet(pfilt, tfilt)

    def UserPropDel(
        self,
        user: str,
        key: str
    ) -> list:
        """Delete a property from a profile

        Args:
            user (str): Name of the profile to delete the property from
            key (str): Name of the property to delete

        Returns:
            list[bool, dict]
        """
        self._RpcClient.UserPropDel(user, [key,])
        user_profile = self._RpcClient.UserPropProfileMultiGet(pfilt=[user,])
        return self._RpcClient.UserPropReplace(
            user,
            user_profile[user]
        )

    def UserPropDelAll(self, user:str) -> None:
        """Deletes a profile from the server

        Args:
            user (str): Profile name to delete
        """
        self._RpcClient.UserPropProfileDelete(user)

    def UserPropCount(self, tfilt: list=None) -> int:
        """Count the number of profiles that exist by filtering on profile type

        Args:
            tfilt (list[str], optional): Profile type to filter by. Defaults to
                None.

        Returns:
            int: Number of profiles that match the given filter
        """
        return self._RpcClient.UserPropProfileCount(tfilt)

    def LocalAuthEnabled(self) -> bool:
        """Check if local authentication is enabled on the remote server

        Returns:
            bool: True if Local Authentication is enabled False otherwise
        """
        return self._RpcClient.LocalAuthEnabled()

    def SetLocalPassword(
        self,
        user: str,
        new_pass: str,
        cur_pass: str=None,
        ignore_checks: bool=False
    ) -> dict:
        """Set the password for a user if using local auth

        This function works a little differently to how the CLI function works
        in that unless checks are specifically ignored, we will enforce them.
        
        Further to the above, the API will not allow a password to be set (when
        checks are not ignored) if there is not already a password set for that
        user.
        Therefore, if we get told there is not a password already set, we will
        perform our own password complexity checks, and send the new password,
        whilst ignoring the server checks, allowing us to set a password for a
        first time/newly created user.

        Raises:
            ApiClientParameterError: current password is None and we're not
                ignoring checks, or API has changed
            ApiClientPasswordIncorrectError: Current password supplied is 
                incorrect
            ApiClientPasswordComplexityError: New password is not complex
                enough
            ApiClientPasswordResetError: Something else happened we didn't
                expect

        Args:
            user (str): user to change password for
            new_pass (str): the new password to set
            cur_pass (str, optional): Current password, only needed if
                ignore_checks is False. If user is new you should provide an
                empty password. Defaults to None.
            ignore_checks (bool, optional): Ignore password complexity and
                current password checks. Defaults to False.
        """
        if not ignore_checks and cur_pass is None:
            raise ApiClientParameterError(
                'Must provide current password if not ignoring password checks'
            )
        
        # Check password complexity before sending to server to save pain
        if not ignore_checks:
            self.is_password_complex(new_pass)

        return_val = self._RpcClient.SetLocalPassword(
            user, new_pass, cur_pass, ignore_checks
        )
        # Check if we are trying to set the password for a new user
        if not return_val['status'] \
            and not ignore_checks \
            and return_val['reason'] == \
            'error verifying current password: no current password is defined':
            # Note that we have already checked the complexity of the password
            # if we are not ignoring checks
            return_val = self._RpcClient.SetLocalPassword(
                user, new_pass, None, True
            )
            
        # Was the current password we sent incorrect?
        if not return_val['status'] \
            and not ignore_checks \
            and return_val['reason'] == \
            ('error verifying current password: failed to enter correct current'
            ' password'):
            raise ApiClientPasswordIncorrectError('Failed to enter the '
                f'correct current password for user "{user}"')

        elif not return_val['status']:
            raise ApiClientPasswordResetError(
                'Something unexpected happened while setting password for '
                f'user "{user}": {return_val["reason"]}'
            )
        else:
            return

    def RemoveLocalPassword(self, user: str) -> None:
        """Remove the password from a user when using local auth

        Args:
            user (str): The user for whom to remove the password
        """
        self._RpcClient.RemoveLocalPassword(user)

    def AutoGenerateClient(self) -> None:
        """Generate a client record for the authenticated user if none exists
        """
        self._RpcClient.AutoGenerateClient()

    def AutoGenerateOnBehalfOf(self, user: str) -> None:
        """Generate a client record for a specific user

        The user passed does not have to exist, and essentially amounts to 
        generating a certificate for a common name equal to the username passed

        Args:
            user (str): The user to create the client record for
        """
        self._RpcClient.AutoGenerateOnBehalfOf(user)
    
    def RevokeCert(self, cn: str) -> None:
        """Revoke a client certificate

        Args:
            cn (str): Common name of the certificate to revoke, usually either
                <username> or <username>_AUTOLOGIN
        """
        self._RpcClient.RevokeCert(cn)

    def RevokeUser(self, user: str) -> None:
        """Revoke all certificates belonging to a given user

        Args:
            user (str): The user whose certificates we want to revoke
        """
        return self._RpcClient.RevokeUser(user)

    def GetUserlogin(self, user: str=None) -> str:
        """Get a user-locked connection profile for the given user

        This is equivalent to Get1(user) except that if the configuration does
        not exist on the server it will be created.

        E.g.

            >>> print(client.Get1('doesnotexist'))
            None
            >>> print(client.GetUserlogin('doesnotexist'))
            # Automatically generated OpenVPN client config file
            . . .
            >>> print(client.Get1('doesnotexist'))
            # Automatically generated OpenVPN client config file
            . . .

        Args:
            user (str, optional): User to get connection profile for. Defaults
                to None

        Returns:
            str: the user-locked configuration profile for the given user, or
                the current authenticated user if None
        """
        return self._RpcClient.GetUserlogin(user)

    def Get1(self, cn: str) -> list:
        """Get a unified connection profile for the given common name

        This connection profile can be written to a file for import into most
        OpenVPN clients. It contains all certs and private keys necessary to
        connect to the server.

        Args:
            cn (str): Common name of the certificate to fetch for the profile

        Returns:
            list[str]: The unified connection profile and name of the file for 
                the given common name or None if it doesn't exist
        """
        return self._RpcClient.Get1(cn)

    def EnumClients(self) -> list:
        """Fetch the list of client names in the database where a client is a
           common name that can connect to the VPN

        Returns:
            list[str]: a list of client names
        """
        return self._RpcClient.EnumClients()

    def ConfigQuery(
        self, prof: str=None, plist: list=None
    ) -> dict:
        """Fetch the configuration by filtering on the given search terms

        Args:
            prof (str, optional): Name of the profile to fetch. Defaults to None
            plist (list[str], optional): List of profile filter terms. These
                are profile entries that we want to extract from the results. 
                Essentially, the columns we select. Defaults to None

        Returns:
            dict[str, str]: Configuration found based on search terms, or
                active config if no search terms defined
        """
        return self._RpcClient.ConfigQuery(prof, plist)

    def DisconnectUser(
        self,
        user: str,
        restart: bool=False,
        reason: str=None,
        client_reason: str=None,
        psid: bool=False
    ) -> int:
        """Disconnect a list of users from the VPN

        Args:
            user (str): The list of users to disconnect from the VPN
            restart (bool, optional): Whether or not to restart the client as
                opposed to just halting them. Defaults to False.
            reason (str, optional): The reason logged on the server for the
                disconnection. Defaults to None.
            client_reason (str, optional): The reason sent to the client for
                disconnection. Defaults to None.
            psid (bool, optional): Whether or not to preserve the most recently
                used SessionID. Use this in combination with restart to invite
                the client to reconnect with the same Session ID. Defaults to
                False.

        Returns:
            int: The number of users kicked, though this should just be 1 or 0
        """
        return self._RpcClient.DisconnectUsers(
            [user,], restart, reason, client_reason, psid
        )

    def Version(self) -> str:
        """Get the version of the server we are communicating with

        Returns:
            str: The version of OpenVPN AccessServer the remote server is 
                running
        """
        return self._RpcClient.GetASLongVersion()

    
    def Status(self) -> dict:
        """Get the run status of the server's internal services

        The returned dictionary contains the following keys:

        * errors (dict): A dictionary of any service errors
        * last_restarted (str): The time the server was last restarted
        * service_status (dict[str, Union["on", "off"]]): status of each 
        internal service

        Below are some of the services that may be listed:

        * api: The XML-RPC API
        * auth: Authentication service (depends on backend? TODO clarify)
        * bridge: Network bridge
        * client_query
        * crl: Certificate Revocation List
        * daemon_pre
        * db_push
        * ip6tables_live
        * ip6tables_openvpn
        * iptables_live
        * iptables_openvpn
        * log
        * openvpn_0
        * openvpn_1
        * subscriptiopn
        * user
        * web

        Returns:
            dict[str, Any]
        """
        return self._RpcClient.RunStatus()

    
    def GetVpnStatus(self) -> dict:
        """Returns a detailed breakdown of the status of the VPN and the client 
        connections

        The resulting dictionary contains a key/value pair for each VPN daemon 
        process. The default configuration likely leaves two open, one 
        internally and the other external for users outside the local network 
        to connect to. They will be called ``openvpn_X`` where 0 is the 
        internal daemon.

        Each of the interface dictionaries contains two lists of import:

        * client_list - a list of connected clients and a status update on them
        * routing_table - a list of all routes currently configured (likely one 
        per client)

        Each of these dictionaries also has a corresponding ``*_header`` 
        dictionary mapping header names to the index of that header within the 
        lists mentioned above.

        Returns:
            dict[str, dict[str, Any]]: A dictionary containing a detailed 
                breakdown of the VPN status
        """
        return self._RpcClient.GetVPNStatus()

    
    def VpnSummary(self) -> dict:
        """Get a summary of the number of clients connected to the VPN

        Returns:
            dict[str, int]: A dictionary with a single key which is 
                ``"n_clients"`` and whose value is the number of clients 
                connected to the VPN
        """
        return self._RpcClient.GetVPNSummary()
    


"""This module provides the UserOperations class which allows us to define 
high-level functionality for managing users on the sacli server
"""
import hashlib
import logging
from typing import Union

import pyovpn_as.api.exceptions

from . import exceptions, utils
from .groups import GroupOperations
from .profile import GroupProfile, ProfileOperations, UserProfile

logger = logging.getLogger(__name__)

class UserOperations(ProfileOperations):
    """Represents the operations we can perform on a given user.

    This class shouldn't be instantiated directly, and instead you should
    access it via AccessServerManagementClient.users.

    This class inherits the methods and attributes from ProfileOperations.

    Args:
        sacli (cli.RemoteSacli): The client we use to communicate with the
            server
    """
    @utils.debug_log_call()
    def get_user(
        self, user: Union[str, UserProfile]
    ) -> UserProfile:
        """Retrieves a user from the server using the specified client

        Args:
            user (Union[str, UserProfile]): The user to fetch

        Raises:
            AccessServerProfileNotFoundError: Profile does not exist
            AccessServerProfileExistsError: Username provided is the name of a
                group, not a user

        Returns:
            UserProfile: A dictionary representing the fetched user
        """
        username = str(user)

        profile = self._get_profile(username)
        
        if profile.type == 'group':
            raise exceptions.AccessServerProfileExistsError(
                f'"{username}" is the name of a group, not a user'
            )
        else:
            return UserProfile(username, profile)


    @utils.debug_log_call(redact=[2, 'password'])
    def create_user(
        self,
        user: Union[str, UserProfile],
        password: str=None,
        group: Union[str, GroupProfile]=None,
        generate_client: bool=True,
        **kwargs
    ) -> UserProfile:
        """Creates a user with the given parameters

        This function will check if the given user exists, and if not will 
        create it via a call to UserPropPut. Then, for every additional 
        parameter set, we will make another call to UserPropPut, deleting the 
        user if any errors occur in the process.

        We can pass a UserProfile if we wish, but any arguments we pass after 
        that will overwrite any properties set in the UserProfile

        Args:
            user (Union[str, UserProfile]): Username of the user to create,
                or a UserProfile object representing the user to create
            password (str, optional): Password to set for user. If None, no
                password will be set. Must adhere to complexity requirements.
                Defaults to None.
            group (Union[str, GroupProfile], optional): A connection group to
                assign the user to. If set, this user will inherit all options 
                that apply to the given group. Will raise an error if group 
                does not exist. Defaults to None. Can be a group profile or a 
                string
            generate_client (bool, optional): Whether or not to generate a
                certificate and openvpn client configuration file for this user.
                Users cannot connect to the VPN unless this is set to true (or a
                cert is generated later for them). Defaults to True. 
            **kwargs:
                prop_superuser (bool, optional): Whether or not to explicitly 
                    make this user an administrator. Defaults to None.
                prop_autologin (bool, optional): Whether or not to explicitly 
                    allow this user to connect without a username and password. 
                    Defaults to None.
                prop_deny (bool, optional): Explicitly deny access to the user. 
                    Defaults to None.
                prop_pwd_change (bool, optional): Whether or not to explicitly 
                    allow the user to change their password using the WebUI. 
                    Defaults to None.
                prop_pwd_strength (bool, optional): Whether or not to
                    explicitly check the complexity of this user's password when
                    they try to change it in the WebUI. Defaults to None.
                prop_autogenerate (bool, optional): Whether or not to prevent 
                    the server from automatically regenerating user connection 
                    profiles (client records) when they don't exist (e.g. 
                    revoked) on the server. Defaults to None

        Raises:
            AccessServerProfileExistsError: username provided already exists as
                either a user or a group
            AccessServerProfileNotFoundError: group given does not exist
            AccessServerConfigError: LocalAuth is not enabled on the server
            ApiClientPasswordComplexityError: Password is not complex enough

        Returns:
            UserProfile: A profile representing the user just created

        TODO Add parameter to hide profile in ui
        TODO explicitly define default behaviour of server in docstring
        TODO Fetch group name from user profile
        TODO tests
        TODO only raise error for local auth if pass not present (in prof too)
        """
        # We're going to be creating a user with a local password
        # Local Auth must therefore be enabled
        if password is not None and not self._sacli.LocalAuthEnabled():
            raise exceptions.AccessServerConfigError(
                'Creating a user with local password requires local auth to be '
                'enabled on the server'
            )

        username = str(user)
        
        # If there is a group specified, check that it exists
        if isinstance(group, GroupProfile):
            group_name = group.group_name
        else:
            group_name = group

        if group_name is not None:
            if not isinstance(group_name, str):
                raise TypeError(
                    f"Expected str for arg 'group', got {type(group_name)}"
                )
            try:
                group_profile = self._get_profile(group_name)
            except exceptions.AccessServerProfileNotFoundError:
                raise exceptions.AccessServerProfileNotFoundError(
                    f'Group "{group_name}" does not exist'
                )
            if not group_profile.is_group:
                raise exceptions.AccessServerProfileExistsError(
                    f'Profile "{group_name}" is not a group'
                )
            logger.debug(f'Got group "{group_name}"')

        # Collect other parameters
        properties = {}
        if group_name is not None:
            properties['conn_group'] = group_name
        property_list = [
            (p, kwargs.get(p)) for p in (
            'prop_superuser',
            'prop_autologin',
            'prop_deny',
            'prop_pwd_change',
            'prop_pwd_strength',
            'prop_autogenerate'
        )]
        for p_name, p_val in property_list:
            if p_val is None:
                continue
            elif not isinstance(p_val, bool):
                raise TypeError(
                    f"Expected bool for arg '{p_name}', got {type(p_val)}"
                )
            elif p_val:
                properties[p_name] = 'true'
            else:
                properties[p_name] = 'false'
        
        # Try to create the user and delete profile if any step fails
        logger.info(f'Creating user "{username}"')
        if isinstance(user, UserProfile):
            new_profile = self._create_profile(username, user, **properties)
        else:
            new_profile = self._create_profile(username, **properties)
        try:
            if password is not None:
                logger.debug(f'Setting password on profile "{username}"')
                try:
                    # Password complexity checked here
                    self._sacli.SetLocalPassword(
                        username, password, ''
                    )
                except pyovpn_as.api.exceptions.ApiClientParameterError \
                as api_err:
                    logger.warning(
                        'Server does not use SetLocalPassword, setting password'
                        ' manually using SHA256 hash'
                    )
                    # Password complexity checks need to be done manually
                    # Remember we need to keep the profile hidden if it is 
                    # already hidden
                    if self._sacli.is_password_complex(password):
                        sha = hashlib.sha256(password.encode())
                        self._sacli.UserPropPut(
                            username, 'pvt_password_digest',
                            sha.hexdigest(), new_profile.is_hidden
                        )
            if generate_client:
                self.create_client_for_user(username)
        except (
            pyovpn_as.api.exceptions.ApiClientBaseException
        ) as api_err:
            logger.error(
                f'Could not create profile "{username}", '
                'aborting and deleting profile...'
            )
            self._sacli.UserPropDelAll(username)
            raise exceptions.AccessServerProfileCreateError(
                'Encountered an issue when setting properties on new user'
            ) from api_err
        else:
            logger.debug(f'Fetching created profile for return...')
            return self.get_user(username)


    @utils.debug_log_call()
    def create_client_for_user(self, user: Union[str, UserProfile]) -> None:
        """Creates a new client record for a given user, or raises an error if 
        one exists

        Args:
            user (Union[str, UserProfile]): User to generate the client for

        Raises:
            AccessServerClientExistsError: A client record for the given user
                already exists
        """
        username = str(user)

        # 1. Verify we are creating a client for an existing user
        self.get_user(username)

        # 2. Check if there is already an existing client
        existing_clients = self._sacli.EnumClients()
        if username in existing_clients:
            raise exceptions.AccessServerClientExistsError(
                f'Client record already exists for "{username}"'
            )
        
        # 3. Create client config and validate it exists
        self._sacli.AutoGenerateOnBehalfOf(username)
        new_existing_clients = self._sacli.EnumClients()
        if username not in new_existing_clients:
            raise exceptions.AccessServerClientCreateError(
                f'Creation of client record for "{username}" failed for an '
                'unknown reason. New client not present on server despite no '
                'returned error'
            )


    @utils.debug_log_call()
    def delete_user(
        self,
        user: Union[str, UserProfile]
    ) -> None:
        """Deletes a user from the server

        Args:
            user (Union[str, UserProfile]): User to delete

        Raises:
            AccessServerProfileNotFoundError: User we are trying to delete does
                not exist
            AccessServerProfileExistsError: username supplied is name of group
                profile
            AccessServerProfileDeleteError: Could not delete the profile for
                an unknown reason
        """
        username = str(user)
        # Check user exists and is a user
        self.get_user(username)
        
        # Delete the user and revoke certs
        self.revoke_user_certificates(username)
        self._delete_profile(username)


    @utils.debug_log_call()
    def get_user_login_ovpn_config(
        self,
        user: Union[str, UserProfile]
    ) -> str:
        """Fetches the .ovpn configuration for 

        Args:
            user (Union[str, UserProfile]): user to fetch configuration for

        Raises:
            AccessServerProfileNotFoundError: User does not have a user/pass
                login connection profile configured or user does not exist.
            AccessServerProfileExistsError: Username provided is the name of a
                group, not a user

        Returns:
            str: The unified connection profile for the given user requiring an
                interactive login
        """
        username = str(user)
            
        # Validate user exists before fetching their config
        self.get_user(username)

        # Use Get1 to prevent config being created in case of not existing
        config = self._sacli.Get1(username)
        if config is None:
            raise exceptions.AccessServerProfileNotFoundError(
                f'Connection profile for user "{username}" could not be found '
                'on the server.'
            )
        else:
            return config[1]


    @utils.debug_log_call()
    def revoke_user_certificates(
        self,
        user: Union[str, UserProfile]
    ) -> None:
        """Revoke all certificates for the given user

        Args:
            user (Union[str, UserProfile]): User whose certificates we want to 
                revoke

        Raises:
            AccessServerProfileNotFoundError: User does not have a user/pass 
                login connection profile configured or user does not exist.
            AccessServerProfileExistsError: Username provided is the name of a
                group, not a user
        """
        username = str(user)

        # Validate user exists
        self.get_user(username)

        # Revoke all certs
        self._sacli.RevokeUser(username)


    @utils.debug_log_call()
    def ban_user(
        self,
        user: Union[str, UserProfile]
    ) -> None:
        """Ban a user from connecting to the VPN

        Args:
            user (Union[str, UserProfile]): The user to ban from the VPN

        Raises:
            AccessServerProfileNotFoundError: User does not exist on the server.
            AccessServerProfileExistsError: Username provided is the name of a
                group, not a user
        """
        username = str(user)
        self.get_user(username)
        self._ban_profile(username)

    
    @utils.debug_log_call()
    def list_users(self) -> list:
        """Lists all users present on the server

        Returns:
            list[UserProfile]: A list of all user profiles on the target server
        """
        profile_dict = self._sacli.UserPropGet(
            tfilt=list(UserProfile.USER_TYPES)
        )
        return [
            UserProfile(user, **props) for user, props in profile_dict.items()
        ]

    
    @utils.debug_log_call()
    def kick_user(
        self,
        user: Union[str, UserProfile],
        reason: str='',
        force: bool=False
    ) -> int:
        """Kick all sessions for a given user and return the number of 
        connections that were kicked

        Args:
            user (Union[str, UserProfile]): The user who we want to kick
            reason (str): The reason to log on the server and to the user for 
                kicking said user
            force (bool): Ban a user immediately after disconnecting them, to 
                prevent clients from attempting to reconnect

        Raises:
            AccessServerProfileNotFoundError: User does not exist on the server.
            AccessServerProfileExistsError: Username provided is the name of a
                group, not a user

        Returns:
            int: The number of connections that were killed
        """
        self.get_user(user)
        username = str(user)
        num_disconnected = self._sacli.DisconnectUser(
            username, reason=reason, client_reason=reason
        )
        if force:
            self.ban_user(username)
        return num_disconnected

    
    @utils.debug_log_call()
    def add_user_to_group(
        self,
        user: Union[str, UserProfile],
        group: Union[str, GroupProfile],
        force_overwrite: bool=False
    ) -> None:
        """Add a given user to the given group by setting the conn_group 
        property on the user

        Args:
            user (Union[str, UserProfile]): User to add to the group
            group (Union[str, GroupProfile]): Group to add the user to
            force_overwrite (bool): Add a user to the new group even if they 
                are already a part of another group

        Raises:
            AccessServerProfileNotFoundError: If either the group or user does 
                not exist
            AccessServerProfileExistsError: If either the group or user names 
                are not names of what they are meant to represent, ie user is 
                actually a group and vice versa
            AccessServerPropOverwriteError: User is already part of a group, 
                and the force_overwrite option is not True
        """
        username = str(user)
        group_name = str(group)

        user_profile = self.get_user(username)
        group_operations = GroupOperations(self._sacli)
        group_operations.get_group(group_name)
        if user_profile.has_group and not force_overwrite:
            raise exceptions.AccessServerPropOverwriteError(
                f"User '{username}' is already part of the group "
                f"'{user_profile.conn_group}'. Remove the user from this "
                "group or call this method with force_overwrite=True"
            )
        elif user_profile.has_group: # and force_overwrite
            logger.warning(
                f"User '{username}' has conn_group='{user_profile.conn_group}',"
                " overwriting with new group..."
            )

        self._sacli.UserPropPut(
            username, 'conn_group', group_name, user_profile.is_hidden
        )


    @utils.debug_log_call()
    def remove_user_from_group(
        self,
        user: Union[str, UserProfile]
    ) -> None:
        """Remove the given user from the group they are a part of

        Args:
            user (Union[str, UserProfile]): The user from which to remove from 
                their group

        Raises:
            AccessServerProfileNotFoundError: User does not exist on the server.
            AccessServerProfileExistsError: Username provided is the name of a
                group, not a user
        """
        username = str(user)
        user_profile = self.get_user(username)
        if not user_profile.has_group:
            logger.debug(f'User not a part of a group, nothing has changed')
        self._sacli.UserPropDel(username, 'conn_group')



import setuptools
from distutils.core import setup

with open('README.md') as readme:
    long_desc = readme.read()

setup(
    name='pyovpn-as',
    packages=[
        'pyovpn_as',
        'pyovpn_as.api'
    ],
    version='0.0.8',
    license='GPL',
    description='A Python library built on XML-RPC that demystifies remote interaction with OpenVPN Access Server',
    long_description=long_desc,
    long_description_content_type='text/markdown',
    author='Peritz',
    author_email='peritz@pardonmynoot.com',
    url='https://github.com/peritz/pyovpn-as',
    download_url='https://github.com/peritz/pyovpn-as/archive/v_0_0_8.tar.gz',
    keywords=['OpenVPN', 'AccessServer', 'Access Server', 'XML-RPC'],
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Libraries',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
    ],
    include_package_data=True
)



"""Contains the classes which represents profiles on the server.

Some notes on profiles:

* If you want to make a group, you can't just pass ``type='group'``, you also have to declare ``group_declare='true'``
* You can't change the type manually, the properties that are set determine what type the record is
* Passing noui as True to UserPropPut does not change anything if
    * ``type`` is ``user_compile`` (due to prop_superuser being set)
    * ``group_declare`` is True
"""
import logging
from typing import Any

import pyovpn_as.api.exceptions
from pyovpn_as.api import cli

from . import exceptions

logger = logging.getLogger(__name__)


class Profile:
    """Represents a profile on the OpenVPN Access Server and provides a logical
    layer to extract meaning from the properties set on the profiles.

    The profile doesn't have to exist on the server for this class to be 
    instantiated, and can be passed as an argument to functions manipulating 
    users, groups, and profiles.

    This class exists primarily as a superclass to User and Group

    Args:
        **attrs (dict[str, Any]): The dictionary containing the attributes for a
            profile. This can be fetched from the server with RemoteSacli
            UserPropGet

    Attributes:
        _attrs (dict[str, Any]): The dictionary containing the attributes for a
            profile provided at ``__init__``
        USER_CONNECT (str): value of ``type`` equal to a profile that is
            evaluated only when a user connects
        USER_CONNECT_HIDDEN (str): value of ``type`` when a profile is
            evaluated the same as USER_CONNECT but is hidden from the admin UI
        USER_COMPILE (str): value of ``type`` when a profile is evaluated on
            iptables compile or when a user connects
        USER_DEFAULT (str): value of ``type`` for the ``__DEFAULT__`` record
        GROUP (str): value of ``type`` for group records
        PROFILE_TYPES (tuple[str]): All types that a profile could be
        USER_TYPES (tuple[str]): List of types valid for a new user
    """
    USER_CONNECT = 'user_connect'
    USER_CONNECT_HIDDEN = 'user_connect_hidden'
    USER_COMPILE = 'user_compile'
    USER_DEFAULT = 'user_default'
    GROUP = 'group'

    PROFILE_TYPES = (
        USER_CONNECT, USER_CONNECT_HIDDEN, USER_COMPILE, USER_DEFAULT, GROUP
    )
    
    USER_TYPES = (
        USER_CONNECT, USER_CONNECT_HIDDEN, USER_COMPILE
    )

    def __init__(self, **attrs):
        for key in attrs:
            if not isinstance(key, str):
                raise TypeError(
                    'All property keys must be strings'
                )
            try:
                attrs[key] = str(attrs[key])
            except:
                raise ValueError('All values must be stringable')
        # Set attributes using Python magic to avoid issues in self.__setattr__
        # We set it twice, once so it is recognised in self.__setattr__ and 
        # again to help with linting
        object.__setattr__(self, '_attrs', {})
        self._attrs = attrs

        # Now force a profile type resolve
        object.__setattr__(self, 'type', self.USER_CONNECT)
        self._resolve_type()


    @property
    def is_hidden(self) -> bool:
        """bool: Whether or not a profile is hidden from the admin interface
        
        True when ``type`` is equal to ``user_connect_hidden``
        """
        return self.type == self.USER_CONNECT_HIDDEN
    

    @property
    def is_banned(self) -> bool:
        """bool: Whether or not the profile is banned.
        
        Derived from the ``prop_deny`` property.

        Default behaviour is False
        """
        prop = self.props.get('prop_deny', '')
        if not isinstance(prop, str):
            raise exceptions.AccessServerProfileIntegrityError(
                f'Type of prop_deny must be str, not a {type(prop)}'
            )
        return prop.lower() == 'true'
    
    
    @property
    def is_admin(self) -> bool:
        """bool: Whether or not the profile is a superuser/admin.
        
        Derived from the ``prop_superuser`` property.

        Default behaviour is False
        """
        prop = self.props.get('prop_superuser', '')
        if not isinstance(prop, str):
            raise exceptions.AccessServerProfileIntegrityError(
                f'Type of prop_superuser must be str, not a {type(prop)}'
            )
        return prop.lower() == 'true'
    

    @property
    def is_group(self) -> bool:
        """bool: Whether or not the profile represents a group. True when the 
        ``group_declare`` property is equal to true
        """
        prop = self.props.get('group_declare', '')
        if not isinstance(prop, str):
            raise exceptions.AccessServerProfileIntegrityError(
                f'Type of group_declare must be str, not a {type(prop)}'
            )
        return prop.lower() == 'true'


    @property
    def can_change_password(self) -> bool:
        """bool: Whether or not users derived from this profile can change
        their password via the web interface.
        
        Derived from the ``prop_pwd_change`` property.

        Default behaviour is False
        """
        prop = self.props.get('prop_pwd_change', '')
        if not isinstance(prop, str):
            raise exceptions.AccessServerProfileIntegrityError(
                f'Type of prop_pwd_change must be str, not a {type(prop)}'
            )
        return prop.lower() == 'true'


    @property
    def can_autologin(self) -> bool:
        """bool: Whether or not users derived from this profile can download a 
        connection profile which allows them to connect without a password.
        
        Derived from the ``prop_autologin`` property.

        Default behaviour is False
        """
        prop = self.props.get('prop_autologin', '')
        if not isinstance(prop, str):
            raise exceptions.AccessServerProfileIntegrityError(
                f'Type of prop_autologin must be str, not a {type(prop)}'
            )
        return prop.lower() == 'true'


    @property
    def will_check_password_strength(self) -> bool:
        """bool: Whether the server should check the password strength when a 
        user derived from this profile tries to change it.
        
        Derived from the ``prop_pwd_strength`` property.

        Default behaviour is True
        """
        prop = self.props.get('prop_pwd_strength', 'true')
        if not isinstance(prop, str):
            raise exceptions.AccessServerProfileIntegrityError(
                f'Type of prop_pwd_strength must be str, not a {type(prop)}'
            )
        return prop.lower() == 'true'


    @property
    def will_autogenerate_client(self) -> bool:
        """bool: Whether or not the server will autogenerate a connection 
        profile for users derived from this profile. If set to true and the 
        given user tries to access their client, the server will generate one 
        if it doesn't exist.
        
        Derived from the ``prop_autogenerate`` property.
        
        Default behaviour is True.
        """
        prop = self.props.get('prop_autogenerate', 'true')
        if not isinstance(prop, str):
            raise exceptions.AccessServerProfileIntegrityError(
                f'Type of prop_autogenerate must be str, not a {type(prop)}'
            )
        return prop.lower() == 'true'

    
    @property
    def props(self) -> dict:
        """dict[str, Any]: The properties set on the profile

        When this is requested, we immediately resolve the type in the case 
        that anything has changed
        """
        self._resolve_type()
        return self._attrs


    def get_prop(self, key: str) -> Any:
        """Get a property from the profile

        Args:
            key (str): The property key to get

        Returns:
            Any: The value at that key

        Raises:
            KeyError: No property defined for that key
        """
        value = self.props.get(
            key, KeyError(
                f"No value for key '{key}' defined."
            )
        )
        if isinstance(value, KeyError):
            raise value
        return value


    def __getattr__(self, attribute: str) -> Any:
        """If we can't get the attribute from regular means, this function is 
        called as a fallback

        In this case we try to fetch the attribute as though it were a property 
        of the profile we were searching for.

        Args:
            attribute (str): The attribute we are searching for

        Returns:
            Any: The value of the attribute from the properties of the profile

        Raises:
            AttributeError: When the property doesn't exist
        """
        value = self.props.get(
            attribute,
            AttributeError(
                f"'{self.__class__.__name__}' object has no attribute "
                f"'{attribute}'"
            )
        )
        if isinstance(value, AttributeError):
            raise value
        return value


    def __setattr__(self, key, value):
        """Sets a property on the profile unless it exists as an attribute on 
        the object. We also evaluate the type of profile we are dealing with

        Args:
            key (str): The property to set
            value (Any): The value of the property
        """
        try:
            self.__getattribute__(key)
        except AttributeError:
            self._attrs[key] = str(value)
            self._resolve_type()
        else:
            object.__setattr__(self, key, value)


    def _resolve_type(self):
        """Resolves the type of profile this is

        Type resolving is based on the following rules:

        * If ``group_declare`` is true, ``type = group``
        * If ``prop_superuser`` is true, ``type = user_compile``
        * If ``inherit``, ``conn_ip``, ``c2s_route``, ``access_from/access_to``, ``dmz_ip``, 
        or ``bypass_route`` are specified, ``type = user_compile``
        * If ``type = user_connect_hidden`` we leave it as is
        * If ``type = user_default`` we leave it as is, but dealing with the 
        default group may cause issues
        * Otherwise type is ``user_connect``
        """
        prof_type = self._attrs.get('type')
        prop_superuser = self._attrs.get('prop_superuser')
        group_declare = self._attrs.get('group_declare')

        if isinstance(group_declare, str) \
            and group_declare.lower() == 'true':
            self._attrs['type'] = self.GROUP
        elif (
            isinstance(prop_superuser, str)
            and prop_superuser.lower() == 'true'
        ) or 'conn_ip' in self._attrs or 'inherit' in self._attrs:
            self._attrs['type'] = self.USER_COMPILE
        else:
            for key in self._attrs:
                if key.startswith((
                    'c2s_route', 'access_from', 'access_to',
                    'dmz_ip', 'bypass_route'
                )):
                    self._attrs['type'] = self.USER_COMPILE
                    break
            if isinstance(prof_type, str) \
                and prof_type.lower() == self.USER_CONNECT_HIDDEN:
                self._attrs['type'] = self.USER_CONNECT_HIDDEN
            if isinstance(prof_type, str) \
                and prof_type.lower() == self.USER_DEFAULT:
                self._attrs['type'] = self.USER_DEFAULT
            else:
                self._attrs['type'] = self.USER_CONNECT   
        self.type = self._attrs['type']     


class UserProfile(Profile):
    """Represents a user's profile on the server.

    This class encapsulates a user's profile on the server (whether or not it 
    exists or not) and provides a logical layer through which we can make sense 
    of a user's properties. This way, we are able to determine if a user is a 
    superuser by checking both their properties and the properties of the group 
    that they are a part of.

    Args:
        username (str): The name of the user whose profile we are representing
        profile (Profile, optional): The profile to derive our attributes from.
            Defaults to None.
        **attrs (dict[str, Any]): The dictionary containing the attributes for a
            profile. This can be fetched from the server with RemoteSacli
            UserPropGet. Any properties defined here will take precedence over 
            those defined in the profile argument
    
    Raises:
        AccessServerProfileIntegrityError: Properties passed do not match a
            UserProfile

    Attributes:
        username (str): The name of the user whose profile we are representing
        _attrs (dict[str, Any]): The dictionary containing the attributes for a
            profile provided at ``__init__``
    """
    def __init__(self, username: str, profile: Profile=None, **attrs):
        if profile is not None and not isinstance(profile, Profile):
            raise TypeError(
                f"Expected 'Profile' for arg 'profile', got {type(profile)}"
            )
        elif profile is not None:
            props = profile._attrs
        else:
            props = {}

        for key, value in attrs.items():
            props[key] = value

        super().__init__(**props)
        if self.type not in self.USER_TYPES:
            raise exceptions.AccessServerProfileIntegrityError(
                f"Properties given do not describe a UserProfile"
            )
        # __setattr__ issues, see Profile class
        object.__setattr__(self, 'username', username)
        self.username = username


    @property
    def has_group(self):
        """bool: Whether or not ``conn_group`` is set"""
        return self.props.get('conn_group') is not None


    def __str__(self):
        """Username of the profile"""
        return self.username

    
    def __setattr__(self, key: str, value: Any):
        """Prevent setting an attribute that would cause the profile to become 
        a group profile

        Args:
            key (str): Name of the attribute to set
            value (Any): Value to set the attribute to

        Raises:
            AccessServerProfileIntegrityError: Tried to set a property that
                would cause the profile to become a group
        """
        if (
            key == 'group_declare'
            and str(value).lower() == 'true'
        ):
            raise exceptions.AccessServerProfileIntegrityError(
                'Attempt made to set group_declare to a true value on a '
                'UserProfile. This is an illegal operation.'
            )
        super().__setattr__(key, value)



class GroupProfile(Profile):
    """Represents a group profile on the server

    This class encapsulates a group and its properties on the server (whether 
    it exists or not) and provides a logical layer through which we can make 
    sense of its properties. We also provide some assurance that the integrity 
    of the group's userprop profile is maintained.

    Args:
        group_name (str): The name of the group whose profile we are
            representing
        profile (Profile, optional): The profile to derive our attributes from.
            Defaults to None.
        **attrs (dict[str, Any]): The dictionary containing the attributes for a
            profile. This can be fetched from the server with RemoteSacli
            UserPropGet. Any properties defined here will take precedence over 
            those defined in the profile argument

    Attributes:
        group_name (str): The name of the group whose profile we are
            representing
        _attrs (dict[str, Any]): The dictionary containing the attributes for a
            profile provided at ``__init__``
    """
    def __init__(self, group_name: str, profile: Profile, **attrs):
        if profile is not None and not isinstance(profile, Profile):
            raise TypeError(
                f"Expected 'Profile' for arg 'profile', got {type(profile)}"
            )
        elif profile is not None:
            props = profile._attrs
        else:
            props = {}

        for key, value in attrs.items():
            props[key] = value

        super().__init__(**props)
        if not self.is_group:
            raise exceptions.AccessServerProfileIntegrityError(
                'Properties given do not describe a GroupProfile'
            )
        # __setattr__ issues, see Profile class
        object.__setattr__(self, 'group_name', group_name)
        self.group_name = group_name


    def __str__(self):
        """Returns the group name of the profile"""
        return self.group_name

    
    def __setattr__(self, key: str, value: Any):
        """Prevent setting an attribute that would cause the profile to become 
        a user profile

        Args:
            key (str): Name of the attribute to set
            value (Any): Value to set the attribute to

        Raises:
            AccessServerProfileIntegrityError: Tried to set a property that
                would cause the profile to become a user
        """
        if (
            key == 'group_declare'
            and str(value).lower() != 'true'
        ):
            raise exceptions.AccessServerProfileIntegrityError(
                'Attempt made to set group_declare to a false value on a '
                'GroupProfile. This is an illegal operation.'
            )
        super().__setattr__(key, value)



class ProfileOperations:
    """A class representing the operations that can take place against a 
    userprop profile.

    This class should not be instantiated directly, and is generally only used 
    to subclass for GroupOperations and UserOperations. This is the reason for 
    making all methods protected.

    Args:
        sacli (cli.RemoteSacli): The client we use to communicate with the
            server
    
    Attributes:
        _sacli (cli.RemoteSacli): The client we use to communicate with the 
            server
    """
    def __init__(self, sacli: cli.RemoteSacli):
        if not isinstance(sacli, cli.RemoteSacli):
            raise TypeError(
                f"Expected 'RemoteSacli' for arg 'sacli', got '{type(sacli)}'"
            )
        self._sacli = sacli


    def _get_profile(self, profile_name: str) -> Profile:
        """Get the userprop profile given by the profile name

        Args:
            profile_name (str): The profile to get from the server

        Raises:
            AccessServerProfileNotFoundError: Profile does not exist

        Returns:
            Profile: Profile representing the userprop profile
        """
        profile_dict = self._sacli.UserPropGet(pfilt=[profile_name,])
        profile = profile_dict.get(profile_name)
        if profile is None:
            raise exceptions.AccessServerProfileNotFoundError(
                f'Could not find profile for "{profile_name}"'
            )
        return Profile(**profile)


    def _create_profile(
        self,
        profile_name: str,
        profile: Profile=None,
        **properties
    ) -> Profile:
        """Create a userprop profile on the server using the profile provided

        Args:
            profile_name (str): Name of the record to insert
            profile (Profile, optional): The profile to create. Default is None
            **properties: Any additional properties to set on the target. These 
                will take precedence over any properties defined in the profile

        Returns:
            Profile: The profile that was created

        Raises:
            AccessServerProfileExistsError: profile provided already exists as
                either a user or a group
        """
        if profile is not None and not isinstance(profile, Profile):
            raise TypeError(
                f"Expected 'Profile' for arg 'profile', got '{type(profile)}'"
            )
        # Check for existence of profile
        try:
            self._get_profile(profile_name)
        except exceptions.AccessServerProfileNotFoundError:
            pass
        else:
            raise exceptions.AccessServerProfileExistsError(
                f'Profile for "{profile_name}" already exists on the server'
            )

        # Resolve the properties we need to add to the profile
        if profile is None:
            new_props = properties
        else:
            new_props = {}
            keys = set(list(properties.keys()) + list(profile.props.keys()))
            for k in keys:
                if k in properties:
                    new_props[k] = properties[k]
                else:
                    new_props[k] = profile.props[k]

        # Create new profile object (checks integrity of attributes)
        new_profile = Profile(**new_props)
        
        try:
            # Set properties required on the new profile
            # Profile will always contain at least the type of user
            for key, value in new_profile.props.items():
                logger.debug(
                    f'Setting property "{key}" on profile "{profile_name}"'
                )
                self._sacli.UserPropPut(
                    profile_name, key, value, new_profile.is_hidden
                )

        except (
            pyovpn_as.api.exceptions.ApiClientBaseException,
            exceptions.AccessServerClientExistsError
        ) as api_err:
            logger.error(
                f'Could not create profile "{profile_name}", '
                'aborting and deleting profile...'
            )
            self._sacli.UserPropDelAll(profile_name)
            raise exceptions.AccessServerProfileCreateError(
                'Encountered an issue when setting properties on new profile'
            ) from api_err
        else:
            logger.debug(f'Fetching created profile for return...')
            return self._get_profile(profile_name)

    
    def _delete_profile(self, profile_name: str) -> None:
        """Delete a given profile from the server

        Args:
            profile_name (str): Profile to delete

        Raises:
            AccessServerProfileNotFoundError: Profile does not exist
            AccessServerProfileDeleteError: Could not delete the profile for
                an unknown reason
        """
        self._sacli.UserPropDelAll(profile_name)

        # Check that the profile is deleted
        try:
            self._get_profile(profile_name)
        except exceptions.AccessServerProfileNotFoundError:
            return
        else:
            raise exceptions.AccessServerProfileDeleteError(
                f'Could not delete profile "{profile_name}" for an unknown '
                'reason'
            )


    def _ban_profile(self, profile_name: str) -> None:
        """Ban users derived from this profile from connecting to the VPN

        Args:
            profile_name (str): Profile to ban

        Raises:
            AccessServerProfileNotFoundError: Profile does not exist
        """
        profile = self._get_profile(profile_name)
        self._sacli.UserPropPut(
            profile_name, 'prop_deny', 'true', profile.is_hidden
        )
